# !/usr/bin/env python3
# Copyright (c) 2024 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import warnings
from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import torch

from quairkit.core import utils

from ..core import get_dtype, tensor_state, to_state, utils
from ..core.intrinsic import (_ArrayLike, _format_system_dim,
                              _is_sample_linear, _SingleParamLike, _StateLike,
                              _type_fetch, _type_transform)
from ..database import pauli_basis

__all__ = [
    "abs_norm",
    "block_enc_herm",
    "create_matrix",
    "dagger",
    "direct_sum",
    "gradient",
    "hessian",
    "herm_transform",
    "kron_power",
    "logm",
    "nkron",
    "NKron",
    "p_norm",
    "partial_trace",
    "partial_trace_discontiguous",
    "partial_transpose",
    "pauli_decomposition",
    "permute_systems",
    "prob_sample",
    "schmidt_decompose",
    "sqrtm",
    "trace",
    "trace_norm",
]


def abs_norm(mat: _StateLike) -> float:
    r"""Tool for calculation of matrix norm.

    Args:
        mat: matrix

    Returns:
        norm of input matrix

    Examples:
        .. code-block:: python

            abs_nor = abs_norm(eye(2) / 2)
            print(f'The abs norm is:\n{abs_nor}')

        ::

            The abs norm is:
            0.7071067690849304
    """
    mat = _type_transform(mat, "tensor")
    mat = mat.to(get_dtype())
    return utils.linalg._abs_norm(mat)


def block_enc_herm(
    mat: _ArrayLike, num_block_qubits: int = 1
) -> _ArrayLike:
    r"""Generate a (qubitized) block encoding of Hermitian ``mat``.

    Args:
        mat: matrix to be block encoded
        num_block_qubits: ancilla qubits used in block encoding

    Returns:
        a unitary that is a block encoding of ``mat``.

    Examples:
        .. code-block:: python

            block_enc = block_enc_herm(x())
            print(f'The block encoding of X is:\n{block_enc}')

        ::

            The block encoding of X is:
            tensor([[0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j],
                    [1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j],
                    [0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j]])
    """
    type_mat = _type_fetch(mat)
    mat = _type_transform(mat, "tensor")

    assert utils.check._is_hermitian(mat), "the input matrix is not a hermitian"
    assert mat.shape[0] == mat.shape[1], "the input matrix is not a square matrix"

    block_enc = utils.linalg._block_enc_herm(mat, num_block_qubits)

    return _type_transform(block_enc, type_mat)


def create_matrix(
    linear_map: Callable[[_ArrayLike], _ArrayLike],
    input_dim: int,
    input_dtype: Optional[torch.dtype] = None,
) -> _ArrayLike:
    r"""Create a matrix representation of a linear map without needing to specify the output dimension.

    This function constructs a matrix representation for a given linear map and input dimension.

    Args:
        linear_map: A function representing the linear map, which takes an input_dim-dimensional vector and returns a vector.
        input_dim: The dimension of the input space.
        input_dtype: The dtype of the input. Defaults to None.

    Returns:
        A matrix representing the linear map.

    Raises:
        RuntimeWarning: the input `linear_map` may not be linear.

    Examples:
        .. code-block:: python

            def f(X):
                return X[0] + X[1]

            mat_repr = create_matrix(f, input_dim=2)
            print(f'The matrix representation is:\n{mat_repr}')

        ::

            The matrix representation is:
            tensor([1.+0.j, 1.+0.j])
    """
    linear_map, generator, input_dtype, type_str = _is_sample_linear(
        func=linear_map, info=[input_dim, 1], input_dtype=input_dtype
    )

    if not utils.check._is_linear(linear_map, generator, input_dtype):
        warnings.warn("the input linear_map may not be linear", RuntimeWarning)

    return _type_transform(
        utils.linalg._create_matrix(
            linear_map=linear_map, input_dim=input_dim, input_dtype=input_dtype
        ),
        type_str,
    )


def dagger(mat: _ArrayLike) -> _ArrayLike:
    r"""Tool for calculation of matrix dagger.

    Args:
        mat: matrix

    Returns:
        The dagger of matrix

    Examples:
        .. code-block:: python

            dag = dagger(t())
            print(f'The dagger of this matrix is:\n{dag}')

        ::

            The dagger of this matrix is:
            tensor([[1.0000-0.0000j, 0.0000-0.0000j],
                    [0.0000-0.0000j, 0.7071-0.7071j]])
    """
    type_mat = _type_fetch(mat)
    mat = _type_transform(mat, "tensor")

    dag = utils.linalg._dagger(mat)

    return _type_transform(dag, type_mat)


def direct_sum(
    A: _ArrayLike, B: _ArrayLike
) -> _ArrayLike:
    r"""Calculate the direct sum of A and B.

    Args:
        A: :math:`m \times n` matrix
        B: :math:`p \times q` matrix

    Returns:
        A direct sum of A and B, with shape :math:`(m + p) \times (n + q)`

    Examples:
        .. code-block:: python

            dir_sum = direct_sum(x(), y())
            print(f'The direct sum is:\n{dir_sum}')

        ::

            The direct sum is:
            tensor([[0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j],
                    [1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 0.+0.j, 0.+0.j, -0.-1.j],
                    [0.+0.j, 0.+0.j, 0.+1.j, 0.+0.j]])
    """
    type_A, type_B = _type_fetch(A), _type_fetch(B)
    A, B = _type_transform(A, "tensor"), _type_transform(B, "tensor")
    B = B.to(A.dtype)

    mat = utils.linalg._direct_sum(A, B)

    return _type_transform(mat, type_A) if type_A == type_B else mat


def gradient(loss_function: Callable[[torch.Tensor], torch.Tensor], var: _ArrayLike, n: int) -> _ArrayLike:
    r"""Compute the gradient of a given loss function with respect to its input variable.

    Args:
        loss_function: A loss function to compute the gradient.
        var: A vector of shape (m, 1) as the input variables for the loss function.
        n: The number of iterations for gradient computation.

    Returns:
        torch.Tensor: The gradient vector of shape (m, 1).

    Examples:
        .. code-block:: python

            def quadratic_loss(x: torch.Tensor) -> torch.Tensor:
                # loss function is: L(x) = x₁² + 2x₂² + 3x₃²
                return x[0]**2 + 2 * x[1]**2 + 3 * x[2]**3

            var = torch.tensor([[1.0], [2.0], [3.0]], requires_grad=True)
            grad = gradient(quadratic_loss, var, n=1)

            print(f"Input variable is:\n{var}")
            print(f"Computed gradient is:\n{grad}")

        ::

            Input variable is:
            tensor([[1.],
                    [2.],
                    [3.]], requires_grad=True)
            Computed gradient is:
            tensor([[ 2.],
                    [ 8.],
                    [81.]], grad_fn=<CopySlices>)
    """
    type_str = _type_fetch(var)
    var = _type_transform(var, "tensor")
    # Check if the input is a vector
    var_ = torch.squeeze(var)
    shape = var_.shape
    if len(shape) != 1:
        warnings.warn(
            f"The input of var is not a (m,1) vector: received {var.shape}", RuntimeWarning
        )
    # Compute and return the gradient
    return _type_transform(utils.linalg._gradient(loss_function, var, n), type_str)


def hessian(loss_function: Callable[[torch.Tensor], torch.Tensor], var: _ArrayLike) -> _ArrayLike:
    r"""Compute the Hessian matrix of a given loss function with respect to its input variables.

    Args:
        loss_function: The loss function to compute the Hessian.
        var: A matrix of shape (n, m) as input variables for the loss function.

    Returns:
        torch.Tensor: Hessian matrix of shape (m, n, n).

    Examples:
        .. code-block:: python

            def quadratic_loss(x: torch.Tensor) -> torch.Tensor:
                # loss function is: L(x) = x₁² + 2x₂² + 3x₃²
                return x[0]**2 + 2 * x[1]**2 + 3 * x[2]**3

            var = torch.tensor([[1.0], [2.0], [3.0]], requires_grad=True)
            hes = hessian(quadratic_loss, var)

            print(f"Input variable is:\n{var}")
            print(f"Computed Hessian is:\n{hes}")

        ::

            Input variable is:
            tensor([[1.],
                    [2.],
                    [3.]], requires_grad=True)
            Computed Hessian is:
            tensor([[[ 2.,  0.,  0.],
                     [ 0.,  4.,  0.],
                     [ 0.,  0., 54.]]])
    """
    shape = var.shape
    # Check if the input is a square matrix
    if len(shape) != 2:
        warnings.warn(
            f"The input of var is not a (n,m) matrix: received {var.shape}", RuntimeWarning
        )
    type_str = _type_fetch(var)
    var = _type_transform(var, "tensor")
    # Compute and return the hessian
    return _type_transform(utils.linalg._hessian(loss_function, var), type_str)


def herm_transform(
    fcn: Callable[[float], float],
    mat: _StateLike,
    ignore_zero: Optional[bool] = False,
) -> _ArrayLike:
    r"""Function transformation for a Hermitian matrix.

    Args:
        fcn: A function :math:`f` that can be expanded by Taylor series.
        mat: Hermitian matrix :math:`H`.
        ignore_zero: Whether to ignore eigenspaces with zero eigenvalue. Defaults to False.

    Returns:
        :math:`f(H)`

    Examples:
        .. code-block:: python

            fH = herm_transform(math.exp, eye(2))
            print(f'The result is:\n{fH}')

        ::

            The result is:
            tensor([[2.7183+0.j, 0.0000+0.j],
                    [0.0000+0.j, 2.7183+0.j]])
    """
    type_str = _type_fetch(mat)
    mat = (
        _type_transform(mat, "tensor")
        if type_str != "state"
        else mat.density_matrix
    )

    assert utils.check._is_hermitian(
        mat
    ), "the input matrix is not Hermitian: check your input"

    mat = utils.linalg._herm_transform(fcn, mat, ignore_zero)

    return mat.detach().numpy() if type_str == "numpy" else mat


def kron_power(matrix: _StateLike, n: int) -> _ArrayLike:
    r"""Calculate the Kronecker product of identical matrices.
    
    Args:
        matrix: The matrix to be powered.
        n: The number of identical matrices.

    Returns:
        Kronecker product of n identical matrices.

    Examples:
        .. code-block:: python

            kp = kron_power(x(), 2)
            print(f'The Kronecker product of 2 X is:\n{kp}')

        ::

            The Kronecker product of 2 X is:
            tensor([[0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j],
                    [0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j],
                    [0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j],
                    [1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j]])
    """
    if n == 0:
        return np.array([[1.0]]) if isinstance(matrix, np.ndarray) else torch.tensor([[1.0]])
    return nkron(matrix, *[matrix for _ in range(n - 1)])


def logm(mat: _StateLike) -> _ArrayLike:
    r"""Calculate the logarithm of a matrix.

    Args:
        mat: Input matrix.

    Returns:
        The matrix of natural base logarithms.
        
    Examples:
        .. code-block:: python

            lgm = logm(x())
            print(f'The log of X is:\n{lgm}')

        ::

            The log of X is:
            tensor([[-1.8562e-16+1.5708j,  1.8562e-16-1.5708j],
                    [ 1.8562e-16-1.5708j, -1.8562e-16+1.5708j]])
    """
    type_str = _type_fetch(mat)
    if type_str == "state":
        return mat.log()

    mat = _type_transform(mat, "tensor")

    return _type_transform(utils.linalg._logm(mat), type_str)


def nkron(matrix_1st: _StateLike, *args: _StateLike) -> _StateLike:
    r"""Calculate the Kronecker product of matrices.

    Args:
        matrix_1st: The first matrix.
        args: Other matrices.

    Returns:
        Kronecker product of the given matrices.

    Examples:
        .. code-block:: python

            A = random_state(1)
            B = random_state(1)
            C = random_state(1)
            result = nkron(A, B, C)
            print(f'The result is:\n{result}')

        ::

            The result is:
            -----------------------------------------------------
             Backend: density_matrix
             System dimension: [2, 2, 2]
             System sequence: [0, 1, 2]
            [[ 0.02+0.j    0.01-0.01j -0.01+0.j   -0.-0.01j  0.04+0.04j  0.05-0.01j
              -0.04-0.03j -0.04+0.02j]
             [ 0.01+0.01j  0.01-0.j   -0.01-0.01j -0.01+0.j   -0.01+0.05j  0.04+0.03j
              -0.-0.04j -0.03-0.02j]
             [-0.01-0.j   -0.01+0.01j  0.02+0.j    0.01-0.02j -0.03-0.04j -0.04+0.j
               0.06+0.06j  0.07-0.02j]
             [-0.-0.01j -0.01-0.j    0.01+0.02j  0.02-0.j    0.02-0.04j -0.02-0.03j
              -0.01+0.08j  0.05+0.05j]
             [ 0.04-0.04j -0.01-0.05j -0.03+0.04j  0.02+0.04j  0.21-0.j    0.1-0.16j
              -0.16+0.03j -0.06+0.14j]
             [ 0.05+0.01j  0.04-0.03j -0.04-0.j   -0.02+0.03j  0.1+0.16j  0.18-0.j
              -0.11-0.11j -0.14+0.03j]
             [-0.04+0.03j -0.-0.04j  0.06-0.06j -0.01-0.08j -0.16-0.03j -0.11+0.11j
               0.29-0.j    0.15-0.23j]
             [-0.04-0.02j -0.03+0.02j  0.07+0.02j  0.05-0.05j -0.06-0.14j -0.14-0.03j
               0.15+0.23j  0.25-0.j  ]]
            -----------------------------------------------------
    """
    if not args:
        return matrix_1st
    
    type_1st = _type_fetch(matrix_1st)
    type_list = [type_1st] + [_type_fetch(arg) for arg in args]

    if all(type_arg == "state" for type_arg in type_list):
        return tensor_state(matrix_1st, *args)

    if all(type_arg == "numpy" for type_arg in type_list):
        return_type = "numpy"
    else:
        return_type = "tensor"

    matrix_1st = _type_transform(matrix_1st, "tensor")
    args = [_type_transform(mat, "tensor") for mat in args]
    result = utils.linalg._nkron(matrix_1st, *args)

    return _type_transform(result, return_type)

NKron = nkron


def p_norm(mat: _StateLike, p: _SingleParamLike) -> _ArrayLike:
    r"""Calculate the Schatten p-norm of a matrix.

    Args:
        mat: matrix
        p: p-norm parameter

    Returns:
        p-norm of input matrix

    Examples:
        .. code-block:: python

            p_nor = p_norm(x(), p=2)
            print(f'The 2-norm of X is:\n{p_nor}')

        ::

            The 2-norm of X is:
            1.4142135381698608
    """
    type_mat = _type_fetch(mat)
    mat = _type_transform(mat, "tensor")
    mat = mat.to(get_dtype())
    p = p if isinstance(p, torch.Tensor) else torch.tensor(p)
    norm = utils.linalg._p_norm_herm(mat, p) if utils.check._is_hermitian(mat) \
        else utils.linalg._p_norm(mat, p)
    return norm.detach().numpy() if type_mat == "numpy" else norm


def partial_trace(
    state: _StateLike, trace_idx: Union[List[int], int], 
    system_dim: Union[List[int], int] = 2
) -> _StateLike:
    r"""Calculate the partial trace of the quantum state.

    Args:
        state: Input quantum state.
        trace_idx: The system indices to be traced out.
        system_dim: The dimension of all systems. Defaults to the qubit case.

    Returns:
        Partial trace of the quantum state with arbitrarily selected subsystem.

    Examples:
        .. code-block:: python

            pt = partial_trace(bell_state(2), 0, [2, 2])
            print(f'The partial trace of Bell state is:\n{pt}')

        ::

            The partial trace of Bell state is:
            -----------------------------------------------------
             Backend: density_matrix
             System dimension: [2]
             System sequence: [0]
            [[0.5+0.j 0. +0.j]
             [0. +0.j 0.5+0.j]]
            -----------------------------------------------------
    """
    type_str = _type_fetch(state)
    if type_str == "state":
        new_state = to_state(state, system_dim)
        return new_state.trace(trace_idx)
    
    trace_idx = [trace_idx] if isinstance(trace_idx, int) else trace_idx
    system_dim = _format_system_dim(state.shape[-1], system_dim)
    assert max(trace_idx) < len(system_dim), \
        f"The trace index {trace_idx} should be smaller than number of systems {len(system_dim)}"
    
    state = _type_transform(state, "state").density_matrix
    state = utils.linalg._partial_trace(state, trace_idx, system_dim)

    return _type_transform(state, type_str)


def partial_trace_discontiguous(
    state: _StateLike, preserve_qubits: List[int] = None
) -> _StateLike:
    r"""Calculate the partial trace of the quantum state with arbitrarily selected subsystem.

    Args:
        state: Input quantum state.
        preserve_qubits: Remaining qubits; if None, all qubits are preserved.

    Returns:
        Partial trace of the quantum state with arbitrarily selected subsystem.

    Examples:
        .. code-block:: python

            ptdis = partial_trace_discontiguous(bell_state(2), [0])
            print(f'The partial trace of Bell state is:\n{ptdis}')

        ::

            The partial trace of Bell state is:
            -----------------------------------------------------
             Backend: density_matrix
             System dimension: [2]
             System sequence: [0]
            [[0.5+0.j 0. +0.j]
             [0. +0.j 0.5+0.j]]
            -----------------------------------------------------
    """
    type_str = _type_fetch(state)
    if type_str == "state":
        new_state = to_state(state, 2)
        return new_state.trace([x for x in range(state.num_qubits) if x not in preserve_qubits])
    
    rho = _type_transform(state, "state", system_dim=2).density_matrix
    
    rho = utils.linalg._partial_trace_discontiguous(rho, preserve_qubits)

    return _type_transform(rho, type_str)


def partial_transpose(
    state: _StateLike, transpose_idx: Union[List[int], int], 
    system_dim: Union[List[int], int] = 2
) -> _ArrayLike:
    r"""Calculate the partial transpose :math:`\rho^{T_A}` of the input quantum state.

    Args:
        state: Input quantum state.
        transpose_idx: The system indices to be transposed.
        system_dim: The dimension of all systems. Defaults to the qubit case.

    Returns:
        The partial transpose of the input quantum state.

    Examples:
        .. code-block:: python

            pt = partial_transpose(bell_state(2), [0])
            print(f'The partial transpose of Bell state is:\n{pt}')

        ::

            The partial transpose of Bell state is:
            -----------------------------------------------------
             Backend: density_matrix
             System dimension: [2, 2]
             System sequence: [0, 1]
            [[0.5+0.j 0. +0.j 0. +0.j 0. +0.j]
             [0. +0.j 0. +0.j 0.5+0.j 0. +0.j]
             [0. +0.j 0.5+0.j 0. +0.j 0. +0.j]
             [0. +0.j 0. +0.j 0. +0.j 0.5+0.j]]
            -----------------------------------------------------
    """
    type_str = _type_fetch(state)
    if type_str == "state":
        new_state = to_state(state, system_dim)
        return new_state.transpose(transpose_idx)
    
    transpose_idx = [transpose_idx] if isinstance(transpose_idx, int) else transpose_idx
    system_dim = _format_system_dim(state.shape[-1], system_dim)
    assert max(transpose_idx) < len(system_dim), \
        f"The transpose index {transpose_idx} should be smaller than number of systems {len(system_dim)}"
    
    type_str = _type_fetch(state)
    state = _type_transform(state, "state").density_matrix
    
    state = utils.linalg._partial_transpose(state, transpose_idx, system_dim)

    return state.detach().numpy() if type_str == "numpy" else state


def pauli_decomposition(mat: _ArrayLike) -> _ArrayLike:
    r"""Decompose the matrix by the Pauli basis.

    Args:
        mat: The matrix to be decomposed.

    Returns:
        A list of coefficients corresponding to the Pauli basis.

    Examples:
        .. code-block:: python

            pauli_dec = pauli_decomposition(random_state(1).density_matrix)
            print(f'The decomposition is:\n{pauli_dec}')

        ::

            The decomposition is:
            tensor([ 0.7071+0.j,  0.5076+0.j,  0.4494+0.j, -0.0572+0.j])
    """
    type_str = _type_fetch(mat)
    mat = _type_transform(mat, "tensor")

    dimension = mat.shape[0]
    num_qubits = int(math.log2(dimension))
    assert (
        2**num_qubits == dimension
    ), f"Input matrix is not a valid quantum data: received shape {mat.shape}"

    basis = pauli_basis(num_qubits)
    coef = torch.cat([torch.trace(mat @ basis[i]).view([1]) for i in range(dimension**2)])
    return _type_transform(coef, type_str)


def permute_systems(
    state: _StateLike, perm_list: List[int], system_dim: Union[List[int], int] = 2,
) -> _StateLike:
    r"""Permute quantum systems based on a permutation list.

    Args:
        state: A matrix representation of a quantum state.
        perm_list: The permutation list. For example, [0, 2, 1, 3] will swap the 2nd and 3rd subsystems.
        system_dim: A list of dimension sizes of each subsystem.

    Returns:
        The permuted matrix.

    Examples:
        .. code-block:: python

            result = permute_systems(random_state(3), [2, 1, 0])
            print(f'The permuted matrix is:\n{result}')

        ::

            The permuted matrix is:
            -----------------------------------------------------
             Backend: density_matrix
             System dimension: [2, 2, 2]
             System sequence: [0, 1, 2]
            [[ 0.06+0.j   -0.02+0.j   -0.02-0.07j -0.03-0.j    0.06+0.01j -0.02+0.03j
              -0.01-0.06j  0.01+0.01j]
             [-0.02-0.j    0.13+0.j   -0.05+0.01j -0.02-0.01j  0.06-0.06j -0.01-0.03j
              -0.+0.08j   0.03-0.03j]
             [-0.02+0.07j -0.05-0.01j  0.24+0.j    0.06+0.03j -0.09+0.05j  0.01-0.05j
               0.04-0.05j -0.13+0.04j]
             [-0.03+0.j   -0.02+0.01j  0.06-0.03j  0.1+0.j    -0.07+0.02j  0.02-0.05j
              -0.01+0.01j -0.03+0.j  ]]
            -----------------------------------------------------
    """
    type_str = _type_fetch(state)
    if type_str == "state":
        new_state = to_state(state, system_dim)
        return new_state.permute(perm_list)
    
    system_dim = _format_system_dim(state.shape[-1], system_dim)
    state = _type_transform(state, "tensor")
    perm_mat = utils.linalg._permute_systems(state, perm_list, system_dim)
    return _type_transform(perm_mat, type_str)


def prob_sample(distribution: _ArrayLike, shots: int = 1024,
                binary: bool = True, proportional: bool = False) -> Dict[str, Union[int, float]]:
    r"""Sample from a probability distribution.

    Args:
        distribution: The probability distribution.
        shots: The number of shots. Defaults to 1024.
        binary: Whether the sampled result is recorded as binary. Defaults to True.
        proportional: Whether the counts are shown in proportion.

    Returns:
        A dictionary containing the ordered sampled results and their counts.

    Examples:
        .. code-block:: python

            dist = torch.abs(haar_state_vector(3))
            result = prob_sample(dist / torch.sum(dist))
            print(f'The sample result is:\n{result}')

        ::

            The sample result is:
            {'0': tensor([1024, 1024, 1024])}
    """
    assert shots > 0, \
        f"The number of shots must be a positive integer, received {shots}"
    distribution = _type_transform(distribution, "tensor")

    prob_sum = distribution.sum(dim=-1, keepdim=True)
    if torch.any(torch.abs(err := prob_sum - 1) > 1e-5):
        warnings.warn(
            "The sum of (some) probability distribution is not close to 1" +
            f" and automatically normalized: received error\n{err}. ", RuntimeWarning
        )
        distribution = distribution / prob_sum
    return utils.linalg._prob_sample(distribution, shots, binary, proportional)


def schmidt_decompose(
    psi: _StateLike, sys_A: List[int] = None
) -> Union[
    Tuple[torch.Tensor, torch.Tensor, torch.Tensor],
    Tuple[np.ndarray, np.ndarray, np.ndarray],
]:
    r"""Calculate the Schmidt decomposition of a quantum state.

    For a state :math:`\lvert\psi\rangle=\sum_i c_i \lvert i_A\rangle\otimes\lvert i_B \rangle`.

    Args:
        psi: State vector form of the quantum state, with shape :math:`(2**n)`.
        sys_A: Qubit indices to be included in subsystem A. By default, the first half of the qubits belong to subsystem A.

    Returns:
        A tuple containing:
            - A one-dimensional array of Schmidt coefficients with shape (k).
            - A high-dimensional array of bases for subsystem A with shape (k, 2**m, 1).
            - A high-dimensional array of bases for subsystem B with shape (k, 2**m, 1).

    Examples:
        .. code-block:: python

            # Example usage (assuming a proper state vector 'psi'):
            c, u, v = schmidt_decompose(psi)
            print("Schmidt coefficients:", c)
            print("Subsystem A bases:", u)
            print("Subsystem B bases:", v)

        ::

            Schmidt coefficients: tensor([...])
            Subsystem A bases: tensor([...])
            Subsystem B bases: tensor([...])
    """
    type_psi = _type_fetch(psi)
    # TODO: Provide a complete example when psi is defined.
    psi = _type_transform(psi, "state_vector").ket

    assert math.log2(
        psi.numel()
    ).is_integer(), "The dimension of input state must be an integral power of 2."

    c, u, v = utils.linalg._schmidt_decompose(psi, sys_A)
    return (
        (
            c.detach().resolve_conj().numpy(),
            u.detach().resolve_conj().numpy(),
            v.detach().resolve_conj().numpy(),
        )
        if type_psi == "numpy"
        else (c, u, v)
    )


def sqrtm(mat: _StateLike) -> _ArrayLike:
    r"""Calculate the square root of a matrix.

    Args:
        mat: Input matrix.

    Returns:
        The square root of the matrix.

    Examples:
        .. code-block:: python

            sqrt = sqrtm(x())
            print(f'The square root of X is:\n{sqrt}')

        ::

            The square root of X is:
            tensor([[0.5000+0.5000j, 0.5000-0.5000j],
                    [0.5000-0.5000j, 0.5000+0.5000j]])
    """
    type_str = _type_fetch(mat)
    if type_str == "state":
        return mat.sqrt()

    mat = _type_transform(mat, "tensor")

    return _type_transform(utils.linalg._sqrtm(mat), type_str)


def trace(mat: _StateLike, axis1: int = -2, axis2: int = -1) -> _ArrayLike:
    r"""Return the sum along the diagonals of the tensor.

    If :math:`mat` is a 2-D tensor, the sum along its diagonal is returned.
    For tensors with more than two dimensions, the axes specified by ``axis1`` and ``axis2``
    determine the 2-D sub-tensors whose traces will be taken.

    Args:
        mat: Input tensor from which the diagonal is taken.
        axis1: The first axis for the 2-D sub-tensor. Defaults to -2.
        axis2: The second axis for the 2-D sub-tensor. Defaults to -1.

    Returns:
        The trace (sum along the diagonal) of the tensor.

    Examples:
        .. code-block:: python

            tr = trace(x())
            print(f'The trace of X is:\n{tr}')

        ::

            The trace of X is:
            0j
    """
    if mat.shape[axis1] != mat.shape[axis2]:
        raise ValueError(
            f"The 2-D tensor from which the diagonal is taken is not square, as {mat.shape[axis1]} != {mat.shape[axis2]}."
        )

    type_str = _type_fetch(mat)
    if type_str == "state":
        assert {axis1, axis2} == {-1, -2}, \
            f"Only support tracing out the state matrix: received {axis1} and {axis2}"
        return mat.trace()
    
    mat_tensor = _type_transform(mat, "tensor")
    trace_mat = utils.linalg._trace(mat_tensor, axis1, axis2)

    return _type_transform(trace_mat, type_str)


def trace_norm(mat: _StateLike) -> _ArrayLike:
    r"""Calculate the trace norm of a matrix.

    Args:
        mat: matrix

    Returns:
        Trace norm of the input matrix

    Examples:
        .. code-block:: python

            tr_norm = trace_norm(x())
            print(f'The trace norm of X is:\n{tr_norm}')

        ::

            The trace norm of X is:
            2.0
    """
    return p_norm(mat, 1)



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Gate matrices.
"""

import itertools
from typing import Callable, List, Optional, Union

import torch

from .. import database
from ..core import get_dtype, utils
from ..core.intrinsic import (_ArrayLike, _ParamLike, _SingleParamLike,
                              _type_fetch, _type_transform)

__all__ = [
    "phase",
    "shift",
    "grover_matrix",
    "qft_matrix",
    "h",
    "s",
    "sdg",
    "t",
    "tdg",
    "eye",
    "x",
    "y",
    "z",
    "p",
    "rx",
    "ry",
    "rz",
    "u3",
    "cnot",
    "cy",
    "cz",
    "swap",
    "cp",
    "crx",
    "cry",
    "crz",
    "cu",
    "rxx",
    "ryy",
    "rzz",
    "ms",
    "cswap",
    "toffoli",
    "universal2",
    "universal3",
    "universal_qudit",
    "Uf",
    "Of",
    "permutation_matrix"
]


def phase(dim: int) -> torch.Tensor:
    r"""Generate phase operator for qudit

    Args:
        dim: dimension of qudit

    Returns:
        Phase operator for qudit

    Examples:
        .. code-block:: python

            dim = 2
            phase_operator = phase(dim)
            print(f'The phase_operator is:\n{phase_operator}')

        ::

            The phase_operator is:
            tensor([[ 1.+0.0000e+00j,  0.+0.0000e+00j],
                    [ 0.+0.0000e+00j, -1.+1.2246e-16j]])
    """
    return utils.matrix._phase(dim, get_dtype())


def shift(dim: int) -> torch.Tensor:
    r"""Generate shift operator for qudit

    Args:
        dim: dimension of qudit

    Returns:
        Shift operator for qudit

    Examples:
        .. code-block:: python

            dim = 2
            shift_operator = shift(dim)
            print(f'The shift_operator is:\n{shift_operator}')

        ::

            The shift_operator is:
            tensor([[0.+0.j, 1.+0.j],
                    [1.+0.j, 0.+0.j]])
    """
    return utils.matrix._shift(dim, get_dtype())


def grover_matrix(oracle: _ArrayLike, dtype: Optional[torch.dtype] = None) -> _ArrayLike:
    r"""Construct the Grover operator based on ``oracle``.

    Args:
        oracle: the input oracle :math:`A` to be rotated.

    Returns:
        Grover operator in form

        .. math::

            G = A (2 |0^n \rangle\langle 0^n| - I^n) A^\dagger \cdot (I - 2|1 \rangle\langle 1|) \otimes I^{n-1}

    Examples:
        .. code-block:: python

            oracle = torch.tensor([[0, 1], [1, 0]], dtype=torch.cfloat)
            grover_op = grover_matrix(oracle)
            print(f'The grover_matrix is:\n{grover_op}')

        ::

            The grover_matrix is:
            tensor([[-1.+0.j,  0.+0.j],
                    [ 0.+0.j, -1.+0.j]])
    """
    oracle = _type_transform(oracle, "tensor")
    return utils.matrix._grover(oracle)


def qft_matrix(num_qubits: int) -> torch.Tensor:
    r"""Construct the quantum Fourier transpose (QFT) gate.

    Args:
        num_qubits: number of qubits :math:`n` st. :math:`N = 2^n`.

    Returns:
        A matrix in the form

        .. math::

            QFT = \frac{1}{\sqrt{N}}
            \begin{bmatrix}
                1 & 1 & \cdots & 1 \\
                1 & \omega_N & \cdots & \omega_N^{N-1} \\
                \vdots & \vdots & \ddots & \vdots \\
                1 & \omega_N^{N-1} & \cdots & \omega_N^{(N-1)^2}
            \end{bmatrix},
            \quad \omega_N = \exp\left(\frac{2\pi i}{N}\right).

    Examples:
        .. code-block:: python

            num_qubits = 1
            qft_gate = qft_matrix(num_qubits)
            print(f'The QFT gate is:\n{qft_gate}')

        ::

            The QFT gate is:
            tensor([[ 0.7071+0.0000e+00j,  0.7071+0.0000e+00j],
                    [ 0.7071+0.0000e+00j, -0.7071+8.6596e-17j]])
    """
    return utils.matrix._qft(num_qubits, get_dtype())


def h() -> torch.Tensor:
    r"""Generate the matrix of the Hadamard gate.

    .. math::

        H = \frac{1}{\sqrt{2}}
            \begin{bmatrix}
                1 & 1 \\
                1 & -1
            \end{bmatrix}

    Returns:
        The matrix of the H gate.

    Examples:
        .. code-block:: python

            H = h()
            print(f'The Hadamard gate is:\n{H}')

        ::

            The Hadamard gate is:
            tensor([[ 0.7071+0.j,  0.7071+0.j],
                    [ 0.7071+0.j, -0.7071+0.j]])
    """
    return utils.matrix._h(get_dtype())


def s() -> torch.Tensor:
    r"""Generate the matrix of the S gate.

    .. math::

        S =
            \begin{bmatrix}
                1 & 0 \\
                0 & i
            \end{bmatrix}

    Returns:
        The matrix of the S gate.

    Examples:
        .. code-block:: python

            S = s()
            print(f'The S gate is:\n{S}')

        ::

            The S gate is:
            tensor([[1.+0.j, 0.+0.j],
                    [0.+0.j, 0.+1.j]])
    """
    return utils.matrix._s(get_dtype())


def sdg() -> torch.Tensor:
    r"""Generate the matrix of the Sdg (S-dagger) gate.

    .. math::

        S^\dagger =
            \begin{bmatrix}
                1 & 0 \\
                0 & -i
            \end{bmatrix}

    Returns:
        The matrix of the Sdg gate.

    Examples:
        .. code-block:: python

            Sdg = sdg()
            print(f'The dagger of S gate is:\n{Sdg}')

        ::

            The dagger of S gate is:
            tensor([[1.+0.j,  0.+0.j],
                    [0.+0.j, -0.-1.j]])
    """
    return utils.matrix._sdg(get_dtype())


def t() -> torch.Tensor:
    r"""Generate the matrix of the T gate.

    .. math::

        T =
            \begin{bmatrix}
                1 & 0 \\
                0 & e^{i\pi/4}
            \end{bmatrix}

    Returns:
        The matrix of the T gate.

    Examples:
        .. code-block:: python

            T = t()
            print(f'The T gate is:\n{T}')

        ::

            The T gate is:
            tensor([[1.0000+0.0000j, 0.0000+0.0000j],
                    [0.0000+0.0000j, 0.7071+0.7071j]])
    """
    return utils.matrix._t(get_dtype())


def tdg() -> torch.Tensor:
    r"""Generate the matrix of the Tdg (T-dagger) gate.

    .. math::

        T^\dagger =
            \begin{bmatrix}
                1 & 0 \\
                0 & e^{-i\pi/4}
            \end{bmatrix}

    Returns:
        The matrix of the Tdg gate.

    Examples:
        .. code-block:: python

            Tdg = tdg()
            print(f'The dagger of T gate is:\n{Tdg}')

        ::

            The dagger of T gate is:
            tensor([[1.0000+0.0000j, 0.0000+0.0000j],
                    [0.0000+0.0000j, 0.7071-0.7071j]])
    """
    return utils.matrix._tdg(get_dtype())


def eye(dim: int = 2) -> torch.Tensor:
    r"""Generate the identity matrix.

    .. math::

        I =
            \begin{bmatrix}
                1 & 0 \\
                0 & 1
            \end{bmatrix}

    Args:
        dim: the dimension of the identity matrix (default is 2 for a qubit).

    Returns:
        The identity matrix.

    Examples:
        .. code-block:: python

            I = eye()
            print(f'The Identity Matrix is:\n{I}')

        ::

            The Identity Matrix is:
            tensor([[1.+0.j, 0.+0.j],
                    [0.+0.j, 1.+0.j]])
    """
    return utils.matrix._eye(dim, get_dtype())


def x() -> torch.Tensor:
    r"""Generate the Pauli X matrix.

    .. math::

        X =
            \begin{bmatrix}
                0 & 1 \\
                1 & 0
            \end{bmatrix}

    Returns:
        The matrix of the X gate.

    Examples:
        .. code-block:: python

            X = x()
            print(f'The Pauli X Matrix is:\n{X}')

        ::

            The Pauli X Matrix is:
            tensor([[0.+0.j, 1.+0.j],
                    [1.+0.j, 0.+0.j]])
    """
    return utils.matrix._x(get_dtype())


def y() -> torch.Tensor:
    r"""Generate the Pauli Y matrix.

    .. math::

        Y =
            \begin{bmatrix}
                0 & -i \\
                i & 0
            \end{bmatrix}

    Returns:
        The matrix of the Y gate.

    Examples:
        .. code-block:: python

            Y = y()
            print(f'The Pauli Y Matrix is:\n{Y}')

        ::

            The Pauli Y Matrix is:
            tensor([[0.+0.j, -0.-1.j],
                    [0.+1.j,  0.+0.j]])
    """
    return utils.matrix._y(get_dtype())


def z() -> torch.Tensor:
    r"""Generate the Pauli Z matrix.

    .. math::

        Z =
            \begin{bmatrix}
                1 & 0 \\
                0 & -1
            \end{bmatrix}

    Returns:
        The matrix of the Z gate.

    Examples:
        .. code-block:: python

            Z = z()
            print(f'The Pauli Z Matrix is:\n{Z}')

        ::

            The Pauli Z Matrix is:
            tensor([[ 1.+0.j,  0.+0.j],
                    [ 0.+0.j, -1.+0.j]])
    """
    return utils.matrix._z(get_dtype())


def p(theta: _SingleParamLike) -> _ArrayLike:
    r"""Generate the P gate matrix.

    .. math::

        P(\theta) =
            \begin{bmatrix}
                1 & 0 \\
                0 & e^{i\theta}
            \end{bmatrix}

    Args:
        theta: the (batched) parameter, with shape [1].

    Returns:
        The matrix of the P gate.

    Examples:
        .. code-block:: python

            theta = torch.tensor([torch.pi / 4])
            p_matrix = p(theta)
            print(f'The P Gate is:\n{p_matrix}')

        ::

            The P Gate is:
            tensor([[1.0000+0.0000j, 0.0000+0.0000j],
                    [0.0000+0.0000j, 0.7071+0.7071j]])
    """
    type_str, theta = _type_fetch(theta), _type_transform(theta, "tensor")
    batch_dim = list(theta.shape[:-1])
    mat = utils.matrix._p(theta).view(batch_dim + [2, 2])
    return _type_transform(mat, type_str)


def rx(theta: _SingleParamLike) -> _ArrayLike:
    r"""Generate the RX gate matrix.

    .. math::

        R_X(\theta) =
            \begin{bmatrix}
                \cos\frac{\theta}{2} & -i\sin\frac{\theta}{2} \\
                -i\sin\frac{\theta}{2} & \cos\frac{\theta}{2}
            \end{bmatrix}

    Args:
        theta: the (batched) parameter, with shape [1].

    Returns:
        The matrix of the RX gate.

    Examples:
        .. code-block:: python

            theta = torch.tensor([torch.pi / 4])
            R_x = rx(theta)
            print(f'The R_x Gate is:\n{R_x}')

        ::

            The R_x Gate is:
            tensor([[0.9239+0.0000j, 0.0000-0.3827j],
                    [0.0000-0.3827j, 0.9239+0.0000j]])
    """
    type_str, theta = _type_fetch(theta), _type_transform(theta, "tensor")
    batch_dim = list(theta.shape[:-1])
    mat = utils.matrix._rx(theta).view(batch_dim + [2, 2])
    return _type_transform(mat, type_str)


def ry(theta: _SingleParamLike) -> _ArrayLike:
    r"""Generate the RY gate matrix.

    .. math::

        R_Y(\theta) =
            \begin{bmatrix}
                \cos\frac{\theta}{2} & -\sin\frac{\theta}{2} \\
                \sin\frac{\theta}{2} & \cos\frac{\theta}{2}
            \end{bmatrix}

    Args:
        theta: the (batched) parameter, with shape [1].

    Returns:
        The matrix of the RY gate.

    Examples:
        .. code-block:: python

            theta = torch.tensor([torch.pi / 4])
            R_y = ry(theta)
            print(f'The R_y Gate is:\n{R_y}')

        ::

            The R_y Gate is:
            tensor([[ 0.9239+0.j, -0.3827+0.j],
                    [ 0.3827+0.j,  0.9239+0.j]])
    """
    type_str, theta = _type_fetch(theta), _type_transform(theta, "tensor")
    batch_dim = list(theta.shape[:-1])
    mat = utils.matrix._ry(theta).view(batch_dim + [2, 2])
    return _type_transform(mat, type_str)


def rz(theta: _SingleParamLike) -> _ArrayLike:
    r"""Generate the RZ gate matrix.

    .. math::

        R_Z(\theta) =
            \begin{bmatrix}
                e^{-i\theta/2} & 0 \\
                0 & e^{i\theta/2}
            \end{bmatrix}

    Args:
        theta: the (batched) parameter, with shape [1].

    Returns:
        The matrix of the RZ gate.

    Examples:
        .. code-block:: python

            theta = torch.tensor([torch.pi / 4])
            R_z = rz(theta)
            print(f'The R_z Gate is:\n{R_z}')

        ::

            The R_z Gate is:
            tensor([[0.9239-0.3827j, 0.0000+0.0000j],
                    [0.0000+0.0000j, 0.9239+0.3827j]])
    """
    type_str, theta = _type_fetch(theta), _type_transform(theta, "tensor")
    batch_dim = list(theta.shape[:-1])
    mat = utils.matrix._rz(theta).view(batch_dim + [2, 2])
    return _type_transform(mat, type_str)


def u3(theta: _ParamLike) -> _ArrayLike:
    r"""Generate the U3 gate matrix.

    .. math::

        U_3(\theta, \phi, \lambda) =
            \begin{bmatrix}
                \cos\frac{\theta}{2} & -e^{i\lambda}\sin\frac{\theta}{2} \\
                e^{i\phi}\sin\frac{\theta}{2} & e^{i(\phi+\lambda)}\cos\frac{\theta}{2}
            \end{bmatrix}

    Args:
        theta: the (batched) parameters, with shape [3, 1].

    Returns:
        The matrix of the U3 gate.

    Examples:
        .. code-block:: python

            theta = torch.tensor([[torch.pi / 4], [torch.pi / 3], [torch.pi / 6]])
            u3_matrix = u3(theta)
            print(f'The U3 Gate is:\n{u3_matrix}')

        ::

            The U3 Gate is:
            tensor([[ 9.2388e-01+0.0000j, -3.3141e-01-0.1913j],
                    [ 1.9134e-01+0.3314j, -4.0384e-08+0.9239j]])
    """
    type_str, theta = _type_fetch(theta), _type_transform(theta, "tensor")
    batch_dim = list(theta.shape[:-1])
    mat = utils.matrix._u3(theta).view(batch_dim + [2, 2])
    return _type_transform(mat, type_str)


def cnot() -> torch.Tensor:
    r"""Generate the CNOT gate matrix.

    .. math::

        CNOT = |0\rangle \langle 0| \otimes I + |1\rangle \langle 1| \otimes X =
            \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 0 & 1 \\
                0 & 0 & 1 & 0
            \end{bmatrix}

    Returns:
        The matrix of the CNOT gate.

    Examples:
        .. code-block:: python

            CNOT = cnot()
            print(f'The CNOT Gate is:\n{CNOT}')

        ::

            The CNOT Gate is:
            tensor([[1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j],
                    [0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j]])
    """
    return utils.matrix._cnot(get_dtype())


def cy() -> torch.Tensor:
    r"""Generate the CY gate matrix.

    .. math::

        CY = |0\rangle\langle0| \otimes I + |1\rangle\langle1| \otimes Y =
            \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 0 & -i \\
                0 & 0 & i & 0
            \end{bmatrix}

    Returns:
        The matrix of the CY gate.

    Examples:
        .. code-block:: python

            CY = cy()
            print(f'The CY Gate is:\n{CY}')

        ::

            The CY Gate is:
            tensor([[1.+0.j,  0.+0.j,  0.+0.j,  0.+0.j],
                    [0.+0.j,  1.+0.j,  0.+0.j,  0.+0.j],
                    [0.+0.j,  0.+0.j,  0.+0.j, -0.-1.j],
                    [0.+0.j,  0.+0.j,  0.+1.j,  0.+0.j]])
    """
    return utils.matrix._cy(get_dtype())


def cz() -> torch.Tensor:
    r"""Generate the CZ gate matrix.

    .. math::

        CZ = |0\rangle\langle0| \otimes I + |1\rangle\langle1| \otimes Z =
            \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 1 & 0 \\
                0 & 0 & 0 & -1
            \end{bmatrix}

    Returns:
        The matrix of the CZ gate.

    Examples:
        .. code-block:: python

            CZ = cz()
            print(f'The CZ Gate is:\n{CZ}')

        ::

            The CZ Gate is:
            tensor([[ 1.+0.j,  0.+0.j,  0.+0.j,  0.+0.j],
                    [ 0.+0.j,  1.+0.j,  0.+0.j,  0.+0.j],
                    [ 0.+0.j,  0.+0.j,  1.+0.j,  0.+0.j],
                    [ 0.+0.j,  0.+0.j,  0.+0.j, -1.+0.j]])
    """
    return utils.matrix._cz(get_dtype())


def swap() -> torch.Tensor:
    r"""Generate the SWAP gate matrix.

    .. math::

        SWAP =
            \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 0 & 1 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 0 & 1
            \end{bmatrix}

    Returns:
        The matrix of the SWAP gate.

    Examples:
        .. code-block:: python

            SWAP = swap()
            print(f'The SWAP Gate is:\n{SWAP}')

        ::

            The SWAP Gate is:
            tensor([[1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j],
                    [0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j]])
    """
    return utils.matrix._swap(get_dtype())


def cp(theta: _SingleParamLike) -> _ArrayLike:
    r"""Generate the CP gate matrix.

    .. math::

        CP(\theta) =
            \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 1 & 0 \\
                0 & 0 & 0 & e^{i\theta}
            \end{bmatrix}

    Args:
        theta: the (batched) parameter, with shape [1].

    Returns:
        The matrix of the CP gate.

    Examples:
        .. code-block:: python

            theta = torch.tensor([torch.pi / 4])
            CP = cp(theta)
            print(f'The CP Gate is:\n{CP}')

        ::

            The CP Gate is:
            tensor([[1.0000+0.0000j, 0.0000+0.0000j, 0.0000+0.0000j, 0.0000+0.0000j],
                    [0.0000+0.0000j, 1.0000+0.0000j, 0.0000+0.0000j, 0.0000+0.0000j],
                    [0.0000+0.0000j, 0.0000+0.0000j, 1.0000+0.0000j, 0.0000+0.0000j],
                    [0.0000+0.0000j, 0.0000+0.0000j, 0.0000+0.0000j, 0.7071+0.7071j]])
    """
    type_str, theta = _type_fetch(theta), _type_transform(theta, "tensor")
    batch_dim = list(theta.shape[:-1])
    mat = utils.matrix._cp(theta).view(batch_dim + [4, 4])
    return _type_transform(mat, type_str)


def crx(theta: _SingleParamLike) -> _ArrayLike:
    r"""Generate the CR_X gate matrix.

    .. math::

        CR_X =
            |0\rangle\langle0| \otimes I + |1\rangle\langle1| \otimes R_X =
            \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & \cos\frac{\theta}{2} & -i\sin\frac{\theta}{2} \\
                0 & 0 & -i\sin\frac{\theta}{2} & \cos\frac{\theta}{2}
            \end{bmatrix}

    Args:
        theta: the (batched) parameter, with shape [1].

    Returns:
        The matrix of the CR_X gate.

    Examples:
        .. code-block:: python

            theta = torch.tensor([torch.pi / 4])
            CR_X = crx(theta)
            print(f'The CR_X Gate is:\n{CR_X}')

        ::

            The CR_X Gate is:
            tensor([[1.0000+0.0000j, 0.+0.0000j, 0.+0.0000j, 0.+0.0000j],
                    [0.+0.0000j, 1.0000+0.0000j, 0.+0.0000j, 0.+0.0000j],
                    [0.+0.0000j, 0.+0.0000j, 0.9239+0.0000j, 0.+-0.3827j],
                    [0.+0.0000j, 0.+0.0000j, 0.+-0.3827j, 0.9239+0.0000j]])
    """
    type_str, theta = _type_fetch(theta), _type_transform(theta, "tensor")
    batch_dim = list(theta.shape[:-1])
    mat = utils.matrix._crx(theta).view(batch_dim + [4, 4])
    return _type_transform(mat, type_str)


def cry(theta: _SingleParamLike) -> _ArrayLike:
    r"""Generate the CR_Y gate matrix.

    .. math::

        CR_Y =
            |0\rangle\langle0| \otimes I + |1\rangle\langle1| \otimes R_Y =
            \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & \cos\frac{\theta}{2} & -\sin\frac{\theta}{2} \\
                0 & 0 & \sin\frac{\theta}{2} & \cos\frac{\theta}{2}
            \end{bmatrix}

    Args:
        theta: the (batched) parameter, with shape [1].

    Returns:
        The matrix of the CR_Y gate.

    Examples:
        .. code-block:: python

            theta = torch.tensor([torch.pi / 4])
            CR_Y = cry(theta)
            print(f'The CR_Y Gate is:\n{CR_Y}')

        ::

            The CR_Y Gate is:
            tensor([[1.0000+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 1.0000+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 0.+0.j, 0.9239+0.j, -0.3827+0.j],
                    [0.+0.j, 0.+0.j, 0.3827+0.j, 0.9239+0.j]])
    """
    type_str, theta = _type_fetch(theta), _type_transform(theta, "tensor")
    batch_dim = list(theta.shape[:-1])
    mat = utils.matrix._cry(theta).view(batch_dim + [4, 4])
    return _type_transform(mat, type_str)


def crz(theta: _SingleParamLike) -> _ArrayLike:
    r"""Generate the CR_Z gate matrix.

    .. math::

        CR_Z =
            |0\rangle\langle0| \otimes I + |1\rangle\langle1| \otimes R_Z =
            \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & e^{-i\theta/2} & 0 \\
                0 & 0 & 0 & e^{i\theta/2}
            \end{bmatrix}

    Args:
        theta: the (batched) parameter, with shape [1].

    Returns:
        The matrix of the CR_Z gate.

    Examples:
        .. code-block:: python

            theta = torch.tensor([torch.pi / 4])
            CR_Z = crz(theta)
            print(f'The CR_Z Gate is:\n{CR_Z}')

        ::

            The CR_Z Gate is:
            tensor([[1.0000+0.0000j, 0.+0.0000j, 0.+0.0000j, 0.+0.0000j],
                    [0.+0.0000j, 1.0000+0.0000j, 0.+0.0000j, 0.+0.0000j],
                    [0.+0.0000j, 0.+0.0000j, 0.9239-0.3827j, 0.+0.0000j],
                    [0.+0.0000j, 0.+0.0000j, 0.+0.0000j, 0.9239+0.3827j]])
    """
    type_str, theta = _type_fetch(theta), _type_transform(theta, "tensor")
    batch_dim = list(theta.shape[:-1])
    mat = utils.matrix._crz(theta).view(batch_dim + [4, 4])
    return _type_transform(mat, type_str)


def cu(theta: _ParamLike) -> _ArrayLike:
    r"""Generate the CU gate matrix.

    .. math::

        CU =
            \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & e^{i\gamma}\cos\frac{\theta}{2} & -e^{i(\lambda+\gamma)}\sin\frac{\theta}{2} \\
                0 & 0 & e^{i(\phi+\gamma)}\sin\frac{\theta}{2} & e^{i(\phi+\lambda+\gamma)}\cos\frac{\theta}{2}
            \end{bmatrix}

    Args:
        theta: the (batched) parameters, with shape [4, 1].

    Returns:
        The matrix of the CU gate.

    Examples:
        .. code-block:: python

            theta = torch.tensor([[torch.pi / 4], [torch.pi / 3], [torch.pi / 6], [torch.pi / 6]])
            CU = cu(theta)
            print(f'The CU Gate is:\n{CU}')

        ::

            The CU Gate is:
            tensor([[1.0000e+00+0.0000j, 0.0000e+00+0.0000j, 0.0000e+00+0.0000j, 0.0000e+00+0.0000j],
                    [0.0000e+00+0.0000j, 1.0000e+00+0.0000j, 0.0000e+00+0.0000j, 0.0000e+00+0.0000j],
                    [0.0000e+00+0.0000j, 0.0000e+00+0.0000j, 8.0010e-01+0.4619j, -1.9134e-01-0.3314j],
                    [0.0000e+00+0.0000j, 0.0000e+00+0.0000j, -1.6728e-08+0.3827j, -4.6194e-01+0.8001j]])
    """
    type_str, theta = _type_fetch(theta), _type_transform(theta, "tensor")
    batch_dim = list(theta.shape[:-1])
    mat = utils.matrix._cu(theta).view(batch_dim + [4, 4])
    return _type_transform(mat, type_str)


def rxx(theta: _SingleParamLike) -> _ArrayLike:
    r"""Generate the RXX gate matrix.

    .. math::

        R_{XX}(\theta) =
            \begin{bmatrix}
                \cos\frac{\theta}{2} & 0 & 0 & -i\sin\frac{\theta}{2} \\
                0 & \cos\frac{\theta}{2} & -i\sin\frac{\theta}{2} & 0 \\
                0 & -i\sin\frac{\theta}{2} & \cos\frac{\theta}{2} & 0 \\
                -i\sin\frac{\theta}{2} & 0 & 0 & \cos\frac{\theta}{2}
            \end{bmatrix}

    Args:
        theta: the (batched) parameter, with shape [1].

    Returns:
        The matrix of the RXX gate.

    Examples:
        .. code-block:: python

            theta = torch.tensor([torch.pi / 4])
            R_XX = rxx(theta)
            print(f'The R_XX Gate is:\n{R_XX}')

        ::

            The R_XX Gate is:
            tensor([[0.9239+0.0000j, 0.0000+0.0000j, 0.0000+0.0000j, 0.0000-0.3827j],
                    [0.0000+0.0000j, 0.9239+0.0000j, 0.0000-0.3827j, 0.0000+0.0000j],
                    [0.0000+0.0000j, 0.0000-0.3827j, 0.9239+0.0000j, 0.0000+0.0000j],
                    [0.0000-0.3827j, 0.0000+0.0000j, 0.0000+0.0000j, 0.9239+0.0000j]])
    """
    type_str, theta = _type_fetch(theta), _type_transform(theta, "tensor")
    batch_dim = list(theta.shape[:-1])
    mat = utils.matrix._rxx(theta).view(batch_dim + [4, 4])
    return _type_transform(mat, type_str)


def ryy(theta: _SingleParamLike) -> _ArrayLike:
    r"""Generate the RYY gate matrix.

    .. math::

        R_{YY}(\theta) =
            \begin{bmatrix}
                \cos\frac{\theta}{2} & 0 & 0 & i\sin\frac{\theta}{2} \\
                0 & \cos\frac{\theta}{2} & -i\sin\frac{\theta}{2} & 0 \\
                0 & -i\sin\frac{\theta}{2} & \cos\frac{\theta}{2} & 0 \\
                i\sin\frac{\theta}{2} & 0 & 0 & \cos\frac{\theta}{2}
            \end{bmatrix}

    Args:
        theta: the (batched) parameter, with shape [1].

    Returns:
        The matrix of the RYY gate.

    Examples:
        .. code-block:: python

            theta = torch.tensor([torch.pi / 4])
            R_YY = ryy(theta)
            print(f'The R_YY Gate is:\n{R_YY}')

        ::

            The R_YY Gate is:
            tensor([[0.9239+0.0000j, 0.0000+0.0000j, 0.0000+0.0000j, 0.0000+0.3827j],
                    [0.0000+0.0000j, 0.9239+0.0000j, 0.0000-0.3827j, 0.0000+0.0000j],
                    [0.0000+0.0000j, 0.0000-0.3827j, 0.9239+0.0000j, 0.0000+0.0000j],
                    [0.0000+0.3827j, 0.0000+0.0000j, 0.0000+0.0000j, 0.9239+0.0000j]])
    """
    type_str, theta = _type_fetch(theta), _type_transform(theta, "tensor")
    batch_dim = list(theta.shape[:-1])
    mat = utils.matrix._ryy(theta).view(batch_dim + [4, 4])
    return _type_transform(mat, type_str)


def rzz(theta: _SingleParamLike) -> _ArrayLike:
    r"""Generate the RZZ gate matrix.

    .. math::

        R_{ZZ}(\theta) =
            \begin{bmatrix}
                e^{-i\theta/2} & 0 & 0 & 0 \\
                0 & e^{i\theta/2} & 0 & 0 \\
                0 & 0 & e^{i\theta/2} & 0 \\
                0 & 0 & 0 & e^{-i\theta/2}
            \end{bmatrix}

    Args:
        theta: the (batched) parameter, with shape [1].

    Returns:
        The matrix of the RZZ gate.

    Examples:
        .. code-block:: python

            theta = torch.tensor([torch.pi / 4])
            R_ZZ = rzz(theta)
            print(f'The R_ZZ Gate is:\n{R_ZZ}')

        ::

            The R_ZZ Gate is:
            tensor([[0.9239-0.3827j, 0.0000+0.0000j, 0.0000+0.0000j, 0.0000+0.0000j],
                    [0.0000+0.0000j, 0.9239+0.3827j, 0.0000+0.0000j, 0.0000+0.0000j],
                    [0.0000+0.0000j, 0.0000+0.0000j, 0.9239+0.3827j, 0.0000+0.0000j],
                    [0.0000+0.0000j, 0.0000+0.0000j, 0.0000+0.0000j, 0.9239-0.3827j]])
    """
    type_str, theta = _type_fetch(theta), _type_transform(theta, "tensor")
    batch_dim = list(theta.shape[:-1])
    mat = utils.matrix._rzz(theta).view(batch_dim + [4, 4])
    return _type_transform(mat, type_str)


def ms() -> torch.Tensor:
    r"""Generate the MS gate matrix.

    .. math::

        MS = R_{XX}\left(-\frac{\pi}{2}\right) = \frac{1}{\sqrt{2}}
            \begin{bmatrix}
                1 & 0 & 0 & i \\
                0 & 1 & i & 0 \\
                0 & i & 1 & 0 \\
                i & 0 & 0 & 1
            \end{bmatrix}

    Returns:
        The matrix of the MS gate.

    Examples:
        .. code-block:: python

            MS = ms()
            print(f'The MS Gate is:\n{MS}')

        ::

            The MS Gate is:
            tensor([[0.7071+0.0000j, 0.0000+0.0000j, 0.0000+0.0000j, 0.0000+0.7071j],
                    [0.0000+0.0000j, 0.7071+0.0000j, 0.0000+0.7071j, 0.0000+0.0000j],
                    [0.0000+0.0000j, 0.0000+0.7071j, 0.7071+0.0000j, 0.0000+0.0000j],
                    [0.0000+0.7071j, 0.0000+0.0000j, 0.0000+0.0000j, 0.7071+0.0000j]])
    """
    return utils.matrix._ms(get_dtype())


def cswap() -> torch.Tensor:
    r"""Generate the CSWAP gate matrix.

    .. math::

        CSWAP =
            \begin{bmatrix}
                1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 \\
                0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 \\
                0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 \\
                0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 \\
                0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 \\
                0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 \\
                0 & 0 & 0 & 0 & 0 & 0 & 0 & 1
            \end{bmatrix}

    Returns:
        The matrix of the CSWAP gate.

    Examples:
        .. code-block:: python

            CSWAP = cswap()
            print(f'The CSWAP Gate is:\n{CSWAP}')

        ::

            The CSWAP Gate is:
            tensor([[1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j],
                    [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j]])
    """
    return utils.matrix._cswap(get_dtype())


def toffoli() -> torch.Tensor:
    r"""Generate the Toffoli gate matrix.

    .. math::

        Toffoli =
            \begin{bmatrix}
                1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 \\
                0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 \\
                0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 \\
                0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 \\
                0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 \\
                0 & 0 & 0 & 0 & 0 & 0 & 0 & 1 \\
                0 & 0 & 0 & 0 & 0 & 0 & 1 & 0
            \end{bmatrix}

    Returns:
        The matrix of the Toffoli gate.

    Examples:
        .. code-block:: python

            Toffoli = toffoli()
            print(f'The Toffoli Gate is:\n{Toffoli}')

        ::

            The Toffoli Gate is:
            tensor([[1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j],
                    [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j]])
    """
    return utils.matrix._toffoli(get_dtype())


def universal2(theta: _ParamLike) -> _ArrayLike:
    r"""Generate the universal two-qubit gate matrix.

    Args:
        theta: the (batched) parameter with shape [15].

    Returns:
        The matrix of the universal two-qubit gate.

    Examples:
        .. code-block:: python

            theta = torch.tensor([
                0.5, 1.0, 1.5, 2.0, 2.5,
                3.0, 3.5, 4.0, 4.5, 5.0,
                5.5, 6.0, 6.5, 7.0, 7.5])
            Universal2 = universal2(theta)
            print(f'The matrix of universal two qubits gate is:\n{Universal2}')

        ::

            The matrix of universal two qubits gate is:
            tensor([[-0.2858-0.0270j,  0.4003+0.3090j, -0.6062+0.0791j,  0.5359+0.0323j],
                    [-0.0894-0.1008j, -0.5804+0.0194j,  0.3156+0.1677j,  0.7090-0.1194j],
                    [-0.8151-0.2697j,  0.2345-0.1841j,  0.3835-0.1154j, -0.0720+0.0918j],
                    [-0.2431+0.3212j, -0.1714+0.5374j,  0.1140+0.5703j, -0.2703+0.3289j]])
    """
    type_str, theta = _type_fetch(theta), _type_transform(theta, "tensor")
    mat = utils.matrix._universal2(theta)
    return _type_transform(mat, type_str)


def universal3(theta: _ParamLike) -> _ArrayLike:
    r"""Generate the universal three-qubit gate matrix.

    Args:
        theta: the (batched) parameter with shape [81].

    Returns:
        The matrix of the universal three-qubit gate.

    Examples:
        .. code-block:: python

            theta = torch.tensor([
                0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0,
                1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2.0,
                2.1, 2.2, 2.3, 2.4, 2.5, 2.6, 2.7, 2.8, 2.9, 3.0,
                3.1, 3.2, 3.3, 3.4, 3.5, 3.6, 3.7, 3.8, 3.9, 4.0,
                4.1, 4.2, 4.3, 4.4, 4.5, 4.6, 4.7, 4.8, 4.9, 5.0,
                5.1, 5.2, 5.3, 5.4, 5.5, 5.6, 5.7, 5.8, 5.9, 6.0,
                6.1, 6.2, 6.3, 6.4, 6.5, 6.6, 6.7, 6.8, 6.9, 7.0,
                7.1, 7.2, 7.3, 7.4, 7.5, 7.6, 7.7, 7.8, 7.9, 8.0,
                8.1])
            Universal3 = universal3(theta)
            print(f'The matrix of universal three qubits gate is:\n{Universal3}')

        ::

            The matrix of universal three qubits gate is:
            tensor([[-0.0675-0.0941j, -0.4602+0.0332j,  0.2635+0.0044j,  0.0825+0.3465j,
                     -0.0874-0.3635j, -0.1177-0.4195j, -0.2735+0.3619j, -0.1760-0.1052j],
                    [ 0.0486+0.0651j, -0.1123+0.0494j,  0.1903+0.0057j, -0.2080+0.2926j,
                     -0.2099+0.0630j, -0.1406+0.5173j, -0.1431-0.3538j, -0.5460-0.1847j],
                    [ 0.0827-0.0303j,  0.1155+0.1111j,  0.5391-0.0701j, -0.4229-0.2655j,
                     -0.1546+0.1943j, -0.0455+0.1744j, -0.3242+0.3539j,  0.3118-0.0041j],
                    [-0.1222+0.3984j,  0.1647-0.1817j,  0.3294-0.1486j, -0.0293-0.1503j,
                      0.0100-0.6481j,  0.2424+0.1575j,  0.2485+0.0232j, -0.1053+0.1873j],
                    [-0.4309-0.0791j, -0.2071-0.0482j, -0.4331+0.0866j, -0.5454-0.1778j,
                     -0.1401-0.0230j,  0.0170+0.0299j,  0.0078+0.2231j, -0.2324+0.3369j],
                    [ 0.0330+0.3056j,  0.2612+0.6464j, -0.2138-0.1748j, -0.2322-0.0598j,
                      0.1387-0.1573j,  0.0914-0.2963j, -0.2712-0.1351j, -0.1272-0.1940j],
                    [ 0.0449-0.3844j,  0.1135+0.2846j, -0.0251+0.3854j,  0.0442-0.0149j,
                     -0.3671-0.1774j,  0.5158+0.1148j,  0.2151+0.1433j, -0.0188-0.3040j],
                    [-0.4124-0.4385j,  0.2306+0.0894j,  0.0104-0.2180j, -0.0180+0.2869j,
                     -0.1030-0.2991j, -0.1473+0.0931j, -0.1686-0.3451j,  0.3825+0.1480j]])
    """
    type_str, theta = _type_fetch(theta), _type_transform(theta, "tensor")
    mat = utils.matrix._universal3(theta)
    return _type_transform(mat, type_str)


def universal_qudit(theta: _ParamLike, dimension: int) -> _ArrayLike:
    r"""Generate a universal gate matrix for qudits using a generalized Gell-Mann basis.

    Args:
        theta: the (batched) parameter with shape [dimension**2 - 1].
        dimension: the dimension of the qudit.

    Returns:
        The matrix of the d-dimensional unitary gate.

    Examples:
        .. code-block:: python

            dimension = 2
            theta = torch.linspace(0.1, 2.5, dimension**2 - 1)
            u_qudit_matrix = universal_qudit(theta, dimension)
            print(f'The matrix of 2-dimensional unitary gate is:\n{u_qudit_matrix}')

        ::

            The matrix of 2-dimensional unitary gate is:
            tensor([[-0.9486+0.2806j,  0.1459+0.0112j],
                    [-0.1459+0.0112j, -0.9486-0.2806j]])
    """
    type_str, theta = _type_fetch(theta), _type_transform(theta, "tensor")
    batch_dim, generator = list(theta.shape[:-1]), database.set.gell_mann(dimension)
    mat = utils.matrix._param_generator(theta, generator).view(batch_dim + [dimension, dimension])
    return _type_transform(mat, type_str)


def Uf(f: Callable[[int], int], n: int) -> torch.Tensor:
    r"""Construct the unitary matrix used in Simon's algorithm.

    Args:
        f: a 2-to-1 or 1-to-1 function :math:`f: \{0, 1\}^n \to \{0, 1\}^n`.
        n: length of the bit string.

    Returns:
        A 2n-qubit unitary matrix satisfying

        .. math::

            U|x, y\rangle = |x, y \oplus f(x)\rangle

    Examples:
        .. code-block:: python

            def f(x: int) -> int:
                return x

            unitary_matrix = Uf(f, 1)
            print(f'Unitary matrix is:\n{unitary_matrix}')

        ::

            Unitary matrix is:
            tensor([[1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j],
                    [0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j],
                    [0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j]])
    """
    d = 2 ** n
    U = torch.zeros(d ** 2, d ** 2)
    for x, y in itertools.product(range(d), range(d)):
        y_out = y ^ f(x)
        input_index = x * d + y
        output_index = x * d + y_out
        U[output_index, input_index] = 1
    return U.to(get_dtype())


def Of(f: Callable[[int], int], n: int) -> torch.Tensor:
    r"""Construct the oracle unitary matrix for an unstructured search problem.

    Args:
        f: a function :math:`f: \{0, 1\}^n \to \{0, 1\}`.
        n: length of the bit string.

    Returns:
        An n-qubit unitary matrix satisfying

        .. math::

            U|x\rangle = (-1)^{f(x)}|x\rangle

    Examples:
        .. code-block:: python

            def f(x: int) -> int:
                return x

            unitary_matrix = Of(f, 1)
            print(f'Unitary matrix is:\n{unitary_matrix}')

        ::

            Unitary matrix is:
            tensor([[ 1.+0.j,  0.+0.j],
                    [ 0.+0.j, -1.+0.j]])
    """
    D = torch.tensor([(-1) ** f(x) for x in range(2 ** n)], dtype=get_dtype())
    return torch.diag(D)


def permutation_matrix(perm: List[int], system_dim: Union[int, List[int]] = 2):
    r"""Construct a unitary matrix representing a permutation operation on a quantum system.

    Args:
        perm: A list representing the permutation of subsystems.
              For example, [1, 0, 2] swaps the first two subsystems.
        system_dim: The dimension of each subsystem.
                    - If an integer, all subsystems are assumed to have the same dimension.
                    - If a list, it specifies the dimension of each subsystem individually.

    Returns:
        A unitary matrix representing the permutation in the Hilbert space.

    Examples:
        .. code-block:: python

            perm = [1, 0]
            U_perm = permutation_matrix(perm, system_dim=2)
            print(f'The permutation matrix is:\n{U_perm}')

        ::

            The permutation matrix is:
            tensor([[0.+0.j, 1.+0.j],
                    [1.+0.j, 0.+0.j]])
    """
    system_dim = [system_dim] * len(perm) if isinstance(system_dim, int) else system_dim
    
    assert len(perm) == len(system_dim), \
        f"perm and system_dim must have the same length, but got perm: {len(perm)} and system_dim: {len(system_dim)}"

    return utils.matrix._permutation(perm, system_dim).to(get_dtype())


for name, func in list(locals().items()):
    if callable(func) and "_" not in name:
        # Add '_gate' to the name for functions without an underscore
        new_name = f"{name}_gate"
        globals()[new_name] = func
        __all__ += [new_name]



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The source file of the class for the measurement.
"""

import math
import warnings
from typing import Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import torch

from ..core import Hamiltonian, Operator, utils
from ..core.intrinsic import _alias, _digit_to_int
from ..core.state import State
from ..database import pauli_str_basis, std_basis

__all__ = ["Measure", "ExpecVal"]


# Debug: torch.trace
class ExpecVal(Operator):
    r"""The class of the loss function to compute the expectation value for the observable.

    This interface can make you using the expectation value for the observable as the loss function.

    Args:
        hamiltonian: The input observable.
        
    .. code-block:: python

        from quairkit.database import random_hamiltonian_generator, random_state

        observable = random_hamiltonian_generator(1)
        print(f'The input observable is:\n{observable}')
        expec_val = ExpecVal(observable)

        input_state = random_state(num_systems=1, rank=2)
        print('The expectation value of the observable:',expec_val(input_state, decompose=True))

        input_state_batch = random_state(num_systems=1, size=2)
        print('The expectation value of the observable:',expec_val(input_state_batch, decompose=False))
                
    ::
    
        The input observable is:
        -0.28233465254251144 Z0
        0.12440505341821817 X0
        -0.2854054036807161 Y0
        The expectation value of the observable: tensor([-0.1162,  0.0768, -0.0081])
        The expectation value of the observable: tensor([0.1748, 0.0198])

    """
    
    def __init__(self, hamiltonian: Hamiltonian):
        super().__init__()
        self.hamiltonian = hamiltonian
    
    def __call__(self, state: State, decompose: Optional[bool] = False) -> Union[torch.Tensor, List[torch.Tensor]]:
        return self.forward(state, decompose)

    def forward(self, state: State, decompose: Optional[bool] = False) -> Union[torch.Tensor, List[torch.Tensor]]:
        r"""Compute the expectation value of the observable with respect to the input state.

        The value computed by this function can be used as a loss function to optimize.

        Args:
            state: The input state which will be used to compute the expectation value.
            decompose: Defaults to ``False``.  If decompose is ``True``, it will return the expectation value of each term.

        Raises:
            NotImplementedError: The backend is wrong or not supported.

        Returns:
            The expectation value.

        """
        return state.expec_val(self.hamiltonian, decompose=decompose)

class Measure(Operator):
    r"""Compute the probability of the specified measurement result.

    Args:
        measure_op: Specify the basis of the measurement. Defaults to ``'computational'``.
        
    Note:
        the allowable input for `measure_op` are:
        - None, i.e., a 'computational' basis
        - a string composed of 'i', 'x', 'y', and 'z'
        - a projection-valued measurement (PVM) in torch.Tensor
    
    .. code-block:: python
    
        from quairkit.database import random_state, std_basis

        # Define measurement basis using a string (e.g., 'xy' denotes eigen-basis of X ⊗ Y)
        op = Measure("xy")
        # Define measurement basis using a list for multiple measurement setups (e.g., 'xy' and 'yz')
        op = Measure(["xy", "yz"])
        # Define a custom measurement basis using a user-specified PVM tensor
        pvm_tensor = std_basis(2).density_matrix
        op = Measure(pvm_tensor)
        # Use default computational basis
        op = Measure()

        state = random_state(num_qubits=2)

        # Full measurement: probability distribution over all measurement outcomes.
        result = op(state)
        print("Probabilities for measuring all qubits:", result)

        # Measurement on a targeted subsystem (e.g., the first qubit)
        result = op(state, system_idx=0)
        print("Probabilities for measuring the first qubit:", result)

        # Compute probability for a specific full-system outcome ('10')
        result = op(state, desired_result='10')
        print("Probability for measuring all qubits with outcome 10:", result)

        # Compute probabilities for selected outcomes (e.g., outcomes corresponding to indices 0 and 3)
        result = op(state, desired_result=[0, 3])
        print("Probabilities for measuring all qubits with outcomes 00 and 11:", result)

        # Retrieve both the probability and the post-measurement state.
        prob, post_state = op(state, keep_state=True)
        print("Post-measurement state:", post_state)

        # Batched measurement on multiple states.
        state_batch = random_state(num_systems=2, size=2)
        result = op(state_batch)
        print(f"Probabilities for measuring two states:\n{result}")
                
    ::
    
        Probabilities for measuring all qubits: tensor([0.1273, 0.0956, 0.3312, 0.4459])
        Probabilities for measuring the first qubit: tensor([0.2229, 0.7771])
        Probability for measuring all qubits with outcome 10: tensor([0.3312])
        Probabilities for measuring all qubits with outcomes 00 and 11: tensor([0.1273, 0.4459])
        Post-measurement state:
        -----------------------------------------------------
        Backend: state_vector
        System dimension: [2, 2]
        System sequence: [0, 1]
        Batch size: [4]

        # 0:
        [1.+0.j 0.+0.j 0.+0.j 0.+0.j]
        # 1:
        [0.+0.j 1.+0.j 0.+0.j 0.+0.j]
        # 2:
        [0.+0.j 0.+0.j 1.+0.j 0.+0.j]
        # 3:
        [0.+0.j 0.+0.j 0.+0.j 1.+0.j]
        -----------------------------------------------------
        
        Probabilities for measuring two states:
        tensor([[0.2846, 0.1826, 0.2665, 0.2663],
                [0.1513, 0.4921, 0.1676, 0.1891]])
    """
    
    def __init__(self, measure_op: Optional[Union[str, List[str], torch.Tensor]] = None) -> None:
        r"""
        
        Note:
            the allowable input for `measure_op` are:
            - None, i.e., a 'computational' basis
            - a string composed of 'i', 'x', 'y', and 'z'
            - a projection-valued measurement (PVM) in torch.Tensor
        """
        super().__init__()

        self.measure_basis = None
        if measure_op is None:
            self.measure_op = measure_op
        
        elif isinstance(measure_op, (str, List)):
            self.measure_basis = pauli_str_basis(measure_op).to(self.dtype)
            self.measure_op = self.measure_basis.density_matrix
        
        elif isinstance(measure_op, torch.Tensor):
            if not all(check := utils.check._is_positive(measure_op).tolist()):
                warnings.warn(
                    f"Some elements of the input PVM is not positive: received {check}", UserWarning)

            sum_basis = torch.sum(measure_op, dim=-3)
            if (err := torch.norm(sum_basis - torch.eye(sum_basis.shape[-1]).expand_as(sum_basis))) > 1e-6:
                warnings.warn(
                    f"The input PVM is not complete: sum distance to identity is {err}", UserWarning)

            #TODO check whether the input is orthogonal
            self.measure_op = measure_op.to(self.dtype)

        else:
            raise ValueError(
                f"Unsupported type for measure_basis: receive type {type(measure_op)}")
    
    @_alias({"system_idx": "qubits_idx"})
    def __call__(self, state: State, system_idx: Optional[Union[Iterable[int], int, str]] = 'full',
                 desired_result: Optional[Union[Iterable[str], str]] = None, keep_state: Optional[bool] = False
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, State]]:
        return self.forward(state, system_idx, desired_result, keep_state)
    
    def __check_measure_op(self, state: State, system_idx: List[int]) -> bool:
        r"""Check whether the shape of input measurement operator is valid,
        and return whether measurement is performed across all qubits.
        """
        system_dim = [state.system_dim[idx] for idx in system_idx]
        measure_op = self.measure_op
        num_measure_systems = len(system_idx)
        
        if measure_op is None:
            self.measure_basis = std_basis(num_measure_systems, system_dim)
            return num_measure_systems == state.num_systems
        
        dim = math.prod(system_dim)
        expected_shape = [dim, dim]
        if measure_op.shape[-2:] != tuple(expected_shape):
            raise ValueError(
                f"The dimension of the PVM does not match the number of qubits: "
                f"received shape {measure_op.shape}, expected {expected_shape}"
            )
        if ((measure_batch_dim := list(measure_op.shape[:-3])) and 
            state._batch_dim and 
            (state._batch_dim != measure_batch_dim)):
            raise ValueError(
                f"The batch dimensions of input state do not match with measurement operator: "
                f"expected None or {measure_batch_dim}, received {state.batch_dim}"
            )
        
        return bool(
            self.measure_basis
            and dim == measure_op.shape[-3]
            and num_measure_systems == state.num_systems
        )

    @_alias({"system_idx": "qubits_idx"})
    def forward(
            self, state: State, system_idx: Optional[Union[Iterable[int], int, str]] = 'full',
            desired_result: Optional[Union[List[Union[int, str]], int, str]] = None, keep_state: Optional[bool] = False
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, State], Tuple[Dict, State]]:
        r"""Compute the probability of measurement to the input state.

        Args:
            state: The quantum state to be measured.
            system_idx: The index of the systems to be measured. Defaults to ``'full'`` which means measure all the qubits.
            desired_result: Specify the results of the measurement to return. Defaults to ``None`` which means return the probability of all the results.
            keep_state: Whether return the measured state. Defaults to ``False``.

        Returns:
            The probability of the measurement.

        """
        num_systems = state.num_systems
        if system_idx == 'full':
            system_idx = list(range(num_systems))
        elif isinstance(system_idx, int):
            system_idx = [system_idx]
        else:
            system_idx = list(system_idx)
        system_dim = [state.system_dim[idx] for idx in system_idx]
        measure_all_sys = self.__check_measure_op(state, system_idx)
        
        prob_array, measured_state = state.measure(self.measure_op, system_idx, keep_state=True)
        
        if measure_all_sys:
            measured_state = self.measure_basis.expand_as(measured_state)
        
        if desired_result or desired_result == 0:
            if isinstance(desired_result, int):
                desired_result = [desired_result]
            elif isinstance(desired_result, str):
                desired_result = [_digit_to_int(desired_result, system_dim)]
            else:
                desired_result = [(_digit_to_int(res, system_dim) if isinstance(res, str) else res)
                                  for res in desired_result]
            
            prob_array = torch.index_select(prob_array, dim=-1, index=torch.tensor(desired_result))
            measured_state = measured_state.prob_select(torch.tensor(desired_result))
        
        return (prob_array, measured_state) if keep_state else prob_array



# !/usr/bin/env python3
# Copyright (c) 2025 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Basic operator.
"""


from typing import Any, Dict, List, Optional, Tuple, TypedDict, Union

import torch

from .base import Backend, get_backend, get_device, get_dtype


class Operator(torch.nn.Module):
    r"""The basic class to implement the operation in QuAIRKit.

    Args:
        backend: The backend implementation of the operator.
            Defaults to ``None``, which means to use the default backend implementation.
        dtype: The data type of the operator.
            Defaults to ``None``, which means to use the default data type.

    """

    def __init__(self, backend: Optional[Backend] = None, dtype: Optional[str] = None):
        if dtype is None:
            self.dtype = get_dtype()
            super().__init__()
        else:
            self.dtype = dtype
            super().__init__(dtype=dtype)
        self.device = get_device()
        self.backend = backend if backend is not None else get_backend()

        self._info: OperatorInfoType = OperatorInfo()

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if isinstance(value, Operator):
            if value.backend is None:
                value.backend = get_backend() if self.backend is None else self.backend
            if value.dtype is None:
                value.dtype = get_dtype() if self.dtype is None else self.dtype

    @property
    def info(self) -> 'OperatorInfoType':
        r"""Information of this operator."""
        return self._info

    def to(self, backend: Optional[Backend] = None, dtype: Optional[str] = None):
        if backend is not None:
            self.backend = backend
            for sub_layer in self.children():
                sub_layer.backend = backend

        if dtype is not None:
            self.dtype = dtype
            for sub_layer in self.children():
                sub_layer.dtype = dtype

    def forward(self, *inputs, **kwargs):
        raise NotImplementedError


class OperatorInfoType(TypedDict):
    r"""A dictionary storing basic information for Operator.

    Keys:
        - name: Operator name.
        - type: Operator type, can be one of:
            - 'general': General operator.
            - 'gate': Quantum gate.
            - 'channel': Quantum channel.
            - 'locc': One-way local operation and classical communication.
        - tex: LaTeX representation (without '$').
        - system_idx: Physical system index or indices the operator acts upon. Shape:
            ``[# of operators, num_acted_system]``
        - num_ctrl_system: Number of indices that control the operator.
        - label: Labels for control systems, collapse events (measurement), or measurement outcomes.
        - param: Parameter tensor for parameterized operators. Shape:
            ``[1 or len(system_idx), batch_size, num_acted_param]``
        - param_sharing: Boolean indicating if parameters are shared. Defaults to ``False``.
        - matrix: Matrix representation (if type is 'gate').
        - api: Circuit API method that invokes the operator in a circuit.
        - permute: Index representation for permutation operators.
        - kwargs: Additional keyword arguments for circuit APIs.
        - plot_width: Plot width (optional, default: ``None``).
    """
    
    name: str
    type: str
    tex: str
    system_idx: List[List[int]]
    num_ctrl_system: int
    label: Union[List[str], List[int], str]
    param: Union[Tuple[float], torch.Tensor]
    param_sharing: bool
    matrix: torch.Tensor
    api: str
    permute: List[int]
    kwargs: Dict[str, Any]

    plot_width: Optional[float]  # TODO depreciated after new matplotlib version


class OperatorInfo(dict):
    __slots__ = ()
    __allowed_keys = {
        "name",
        "type",
        "tex",
        "system_idx",
        "num_ctrl_system",
        "label",
        "param",
        "param_sharing",
        "matrix",
        "api",
        "permute",
        "kwargs",
        "plot_width",  # TODO depreciated after new matplotlib version
    }
    __open_qasm: Dict[str, Optional[str]] = {
        'x': 'x',
        'ctrl-x': 'cx',
        'rx': 'rx',
        'oracle': None,
    }
    r"""A dictionary of translating QuAIRKit operators to 
    [QpenQASM 3 operators](https://openqasm.com/language/standard_library.html#standard-library)
    
    Here value None means that the operator is not supported in OpenQASM 3.
    
    TODO: Add more operators.
    """

    def __init__(self, name: str = "unidentified", type: str = "general"):
        super().__init__(
            name=name,
            type=type,
        )

    def __setitem__(self, key, value):
        if key not in self.__allowed_keys:
            raise KeyError(
                f"Invalid key '{key}' for OperatorInfo, accept one of {self.__allowed_keys}."
            )
        super().__setitem__(key, value)

    def update(self, *args, **kwargs):
        for key, value in dict(*args, **kwargs).items():
            self[key] = value

    @property
    def qasm(self) -> str:
        r"""Translate to open qasm format
        
        TODO: incorporate with __open_qasm
        """
        name = self['name']
        list_system_idx, idx_str = self['system_idx'], ''
        for system_idx in list_system_idx:
            idx_str += f"{name} q[{system_idx[0]}]"
            for idx in system_idx[1:]:
                idx_str += f", q[{idx}]"
            idx_str += ';\n'
        return idx_str



# !/usr/bin/env python3
# Copyright (c) 2025 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Source file for transforming OperatorInfo to string representing quantikz.

Quantikz is a LaTeX package for typesetting quantum circuit diagrams.
See https://ctan.org/pkg/quantikz for more details.
"""

import datetime
import os
import subprocess
import warnings
from copy import deepcopy
from typing import Dict, List, Optional, Tuple, Union

import IPython
import numpy as np
import torch
from pdf2image import convert_from_path

from .operator import OperatorInfoType

__all__ = ['code_to_str', 'OperatorListDrawer']


def _format_line(list_s: List[str], begin_code: str, next_line: bool = True) -> str:
    r"""Format a list of strings into a LaTeX line, connected by & (and \\ if contain next line)
    """
    begin_code = f"{begin_code} & " if begin_code else "& "
    return begin_code + " & ".join(list_s) + (r" \\" +  "\n" if next_line else "")

def code_to_str(code: Dict[int, List[str]], 
               begin_code: Optional[Dict[int, str]] = None) -> str:
    r"""Translate the given quantikz code into complete LaTeX commands
    
    Args:
        code: The quantikz code to translate, where the key i is the i-th line of the code, 
                and the value is the list of gate commands on that line.
        begin_code: a dictionary or a string representing the beginning code for each line. 
                Defaults to empty.
        
    Returns:
        The complete LaTeX code for the given quantikz code, connected by & and \\
    
    """
    system_idx = sorted(code.keys())
    assert len(system_idx) == system_idx[-1] + 1, \
        f"The system indices must be consecutive and start from 0: received {system_idx}"
    
    begin_code = {} if begin_code is None else begin_code
    last_line = _format_line(code[system_idx[-1]], begin_code.get(system_idx[-1], ''), next_line=False)
    other_lines = ''.join(_format_line(code[idx], begin_code.get(idx, '')) for idx in system_idx[:-1])
    return ''.join(other_lines) + last_line


def _format_number(num: Union[int, float, complex], decimals: int, simplify_zero: bool = False) -> str:
    if simplify_zero and np.abs(num) < (10 ** -decimals):
        return '0'
    return f"{num:.{decimals}f}".replace('(', '').replace(')', '').replace('j', 'i')

def _format_vector(data: np.ndarray, decimals: int) -> str:
    present_data = data[:3]
    row_str = [_format_number(x, decimals) for x in present_data]
    if len(data) > 3:
        row_str.append(r'\ldots')
    return r', '.join(row_str)

def _format_matrix(data: np.ndarray, decimals: int) -> str:
    assert len(data.shape) == 2, \
        f"Only 2D arrays are supported: received {data.shape}"
    
    text = r'\begin{bmatrix}' + '\n'
    for x in range(len(data)):
        row_str = [_format_number(data[x, y], decimals, simplify_zero=True) for y in range(len(data[x]))]
        text += r' & '.join(row_str)
        if x != len(data) - 1:
            text += r' \\' + '\n'
    text += '\n' + r'\end{bmatrix}'
    return text

def _format_data(data: torch.Tensor, decimals: int) -> str:
    r"""Format the given data tensor into flatten LaTex numbers or LaTeX matrix
    
    Args:
        data: The data tensor to format
        decimals: The number of decimals to show
        to_matrix: Whether to format the data into a 2D matrix
    
    """
    data = np.round(data.squeeze().detach().cpu().numpy(), decimals=decimals)
    
    if len(data.shape) == 0:
        return _format_number(data, decimals)
    
    elif len(data.shape) == 1:
        return _format_vector(data, decimals)
    
    elif len(data.shape) == 2:
        return _format_matrix(data, decimals)
    
    raise ValueError(
        f"Only 1D and 2D tensors are supported: received {data.shape}")


def _permute_str(perm: List[int], system_idx: List[int]) -> Tuple[str, Tuple[List[int], List[int]]]:
    r"""Give the quantikz permutation of the systems
    
    Args:
        perm: permutation
        system_idx: systems to permute
    
    """
    assert len(perm) == len(system_idx), \
        f"Lengths of permutation and system_idx mismatch: received {perm} and {system_idx}"
    min_idx = min(system_idx)
    perm_map = {system_idx[i]: system_idx[perm[i]] for i in range(len(perm))}
    
    for idx in range(min_idx, max(system_idx) + 1):
        perm_map.setdefault(idx, idx)
    target_param = [str(perm_map[idx] - min_idx + 1) for idx in sorted(perm_map)]
    return fr"\permute{{{','.join(target_param)}}}"


def __clear_temp_files(file_name: str) -> None:
    for ext in [".aux", ".log", ".pdf", ".tex", ".png"]:
        temp_file = f"{file_name}{ext}"
        if os.path.exists(temp_file):
            os.remove(temp_file)

def _plot_code(latex_code: str, dpi: int) -> IPython.display.Image:
    r"""Plot the given LaTeX code using local LaTeX distribution
    """
    # Write the LaTeX code to a file
    date = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
    file_name = f"quairkit{date}temp"
    with open(f"{file_name}.tex", "w") as f:
        f.write(latex_code)

    # Run pdflatex to compile the TeX file
    result = subprocess.run(
        ["pdflatex", "-interaction=nonstopmode", f"{file_name}.tex"],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL
    )

    if result.returncode != 0:
        log_path = f"{file_name}.log"
        log_output = ""
        if os.path.exists(log_path):
            with open(log_path, "r") as log_file:
                log_output = log_file.read()
            print("LaTeX compilation failed. Log output:")
            print(log_output)
        else:
            print("Compilation failed, and no log file was generated.")
            
        __clear_temp_files(file_name)
        raise RuntimeError("LaTeX compilation failed. See log output for details.")

    # Convert the resulting PDF to an image using pdf2image
    pdf_path = f"{file_name}.pdf"
    try:
        images = convert_from_path(pdf_path, dpi=dpi)
        images[0].save(f"{file_name}.png", "PNG")
    except Exception as e:
        __clear_temp_files(file_name)
        raise RuntimeError(f"Failed to convert PDF to PNG: {e}") from e

    obj = IPython.display.Image(filename=f"{file_name}.png")
    __clear_temp_files(file_name)
    return obj


class OperatorListDrawer:
    r"""Latex plot for a list of operators in QuAIRKit
    
    Args:
        style: the plot style of the circuit, can be 'standard', 'compact' or 'detailed'.
    
    Note:
        The LaTeX package used is [quantikz](https://ctan.org/pkg/quantikz).
    
    """
    def __init__(self, style: str, decimals: int) -> None:
        self.style = style
        self.decimals = decimals
        self._code: Dict[int, List[str]] = {}
        
    def _fill_empty(self, system_idx: List[int], classical_idx: Optional[List[int]] = None) -> None:
        r"""Fill empty &s for incoming multi-system operators
        """
        min_idx, max_idx = min(system_idx), max(system_idx)
        if min_idx == max_idx:
            self._code.setdefault(min_idx, [])
            return
        
        lengths = {}
        for idx in range(min_idx, max_idx + 1):
            current_str = self._code.setdefault(idx, [])
            lengths[idx] = len(current_str)
        max_len = max(lengths.values()) + 1
        
        # Pad each line up to max_len
        classical_idx = classical_idx if classical_idx is not None else []
        for idx, old_len in lengths.items():
            pad_amount = max_len - (old_len + 1)
            
            if idx in classical_idx:
                self._code[idx] += [r"\wireoverride{c}"] * pad_amount
            else:
                self._code[idx] += [r"{}"] * pad_amount
            
    def fill_all(self) -> None:
        r"""Fill empty &s for all lines covered by the current codes
        """
        min_idx, max_idx = min(self._code.keys()), max(self._code.keys())
        self._fill_empty(list(range(min_idx, max_idx + 1)))
            
    def clone(self) -> 'OperatorListDrawer':
        r"""Clone the current drawer
        """
        new_drawer = OperatorListDrawer(self.style, self.decimals)
        new_drawer._code = deepcopy(self._code)
        return new_drawer
        
    @property
    def code(self) -> Dict[int, List[str]]:
        r"""The dictionary of the current circuit code
        
        Note:
            Empty lines are filled with '{}'
        """
        if not list(self._code.keys()):
            return {}
        self.fill_all()
        return deepcopy(self._code)
    
    def add_end(self, rstick_str: Optional[Dict[int, str]] = None) -> None:
        r"""Add end sticks to the circuit
        
        Args:
            rstick_str: The string to add to the right stick at idx, such as \ket{0}
        
        """
        rstick_str = {} if rstick_str is None else rstick_str
        for idx in self._code:
            self._code[idx].append(r"\rstick{" + rstick_str.get(idx, '') +  r'}')
    
    __plot_begin_code = r'''\documentclass[border=2pt]{standalone}

% WARNING: please disable other quantum circuit packages, such as qcircuit
\usepackage{tikz}
\usetikzlibrary{quantikz2}

\begin{document}

\begin{quantikz}[transparent]
'''
    
    __end_code = r'''
\end{quantikz}

\end{document}
'''
    
    def plot(self, dpi: int = 300, print_code: bool = True, 
             begin_code: Optional[Dict[int, str]] = None) -> IPython.display.Image:
        r"""Plot the given quantikz code using LaTeX
        
        Args:
            dpi: The DPI to use for the resulting image. Defaults to 300.
            print_code: Whether to print the code to the console. Defaults to True.
            begin_code: a dictionary or a string representing the beginning code for each line. 
                Defaults to empty.
            
        Returns:
            The image object of the plot
        
        """
        if begin_code is None:
            begin_code = {idx: r'\lstick{}' for idx in self._code.keys()}
        latex_code = self.__plot_begin_code + code_to_str(self.code, begin_code) + self.__end_code
        
        if print_code:
            print(latex_code)
        return _plot_code(latex_code, dpi)
    
    def _catenate(self, drawer: 'OperatorListDrawer') -> 'OperatorListDrawer':
        r"""Catenate the given drawers
        
        Args:
            *drawers: The drawers to catenate
        
        """
        if not list(drawer._code.keys()):
            return self.clone()
        if not list(self._code.keys()):
            return drawer.clone()
        
        new_drawer = self.clone()
        
        if list(set(new_drawer._code.keys()) & set(drawer._code.keys())) :
            new_drawer._fill_empty(list(drawer._code.keys()))
            
            for idx, code in drawer._code.items():
                new_drawer._code[idx] = new_drawer._code.get(idx, []) + code
        else:
            for idx, code in drawer._code.items():
                new_drawer._code[idx] = code
        return new_drawer
    
    def __add__(self, drawer: 'OperatorListDrawer') -> 'OperatorListDrawer':
        return self._catenate(drawer)
    
    def _add_ctrl(self, system_idx: List[int], num_ctrl_system: int, control_label: str, classical: bool = False) -> Tuple[int, int]:
        r"""Add control lines to the given system index
        
        Returns:
            The minimum and maximum system index the input operator applies
        
        """
        if num_ctrl_system == 0:
            return min(system_idx), max(system_idx)
        assert control_label and len(control_label) == num_ctrl_system, \
                f"Control label must be provided for each control system: received {control_label} with {num_ctrl_system} control systems"
        ctrl_system_idx, apply_system_idx = system_idx[:num_ctrl_system], system_idx[num_ctrl_system:]

        min_apply_idx, max_apply_idx = min(apply_system_idx), max(apply_system_idx)
        min_ctrl_idx, max_ctrl_idx = min(ctrl_system_idx), max(ctrl_system_idx)

        if min_ctrl_idx > max_apply_idx:
            start_idx, end_idx = max_ctrl_idx, max_apply_idx
            min_idx, max_idx = max_apply_idx + 1, max_ctrl_idx
        elif max_ctrl_idx < min_apply_idx:
            start_idx, end_idx = min_ctrl_idx, min_apply_idx
            min_idx, max_idx = min_ctrl_idx, min_apply_idx - 1
        else:
            raise ValueError(
                f"Control systems and apply systems must be separated: received {ctrl_system_idx} controlling {apply_system_idx}")
        across_system = end_idx - start_idx

        classical_prefix = r"vertical wire=c" if classical else ''

        for idx in range(min_idx, max_idx + 1):
            if idx == start_idx: # the first control system
                label = control_label[ctrl_system_idx.index(idx)]
                if label == '0':
                    self._code[idx].append(r"\octrl[" + classical_prefix + r"]{" + str(across_system) + r'}')
                elif label == '1':
                    self._code[idx].append(r"\ctrl[" + classical_prefix + r"]{" + str(across_system) + r'}')
                else:
                    self._code[idx].append(r"\ctrl[style={fill=white,draw=black,inner sep=0.3pt}," + classical_prefix + 
                                           r"]{" + str(across_system) + r"} \push{\scriptscriptstyle " + label + r"}")
            elif idx in ctrl_system_idx:
                label = control_label[ctrl_system_idx.index(idx)]
                if label == '0':
                    self._code[idx].append(r"\control[style={fill=white,draw=black,inner sep=1pt}]{}")
                elif label == '1':
                    self._code[idx].append(r"\control{}")
                else:
                    self._code[idx].append(r"\control[style={fill=white,draw=black,inner sep=0.3pt}]{} \push{\scriptscriptstyle " + 
                                           label + r"}")
            else:
                self._code[idx].append(r"\push{\,\,}")

        return min_apply_idx, max_apply_idx
    
    __empty_through = (r'\gateinput[label style={xshift=-8.2pt}]{$\textcolor{white}{\rule{4pt}{4pt}}$}' + 
                       r'\gateoutput[label style={xshift=8.2pt}]{$\textcolor{white}{\rule{4pt}{4pt}}$}')
        
    def _append_general_op(self, info: OperatorInfoType) -> None:
        r"""Append a general fixed operator
        
        Note:
            Presentation rule is as follows:
            - If the style is 'detailed', the matrix is provided and not controlled, show the matrix;
            - Otherwise, just show the name.
        
        """
        info.setdefault("tex", fr"\operatorname{{{info['name']}}}")
        num_ctrl_system, control_label = info.get("num_ctrl_system", 0), info.get("label", None)
        show_matrix = (
            self.style == "detailed"
            and ("matrix" in info and len(info["matrix"].shape) <= 2 and info["matrix"].shape[-1] <= 10)
            and num_ctrl_system == 0
        )
        _empty_through = r"{}" if "permute" in info else self.__empty_through
        
        for idx, system_idx in enumerate(info["system_idx"]):
            self._fill_empty(system_idx)
            min_idx, max_idx = self._add_ctrl(system_idx, num_ctrl_system, control_label)
            num_across_system = max_idx - min_idx + 1
            
            if show_matrix:
                matrix = info["matrix"]
                if len(matrix.shape) > 2:
                    matrix = matrix[idx]
                gate_name = _format_data(matrix, self.decimals)
            elif "permute" in info:
                gate_name = _permute_str(info["permute"], system_idx[num_ctrl_system:])
            else:
                gate_name = info["tex"]
            
            if gate_name == r"\targ{}":
                gate_command = gate_name
            else:
                gate_prefix = str(num_across_system)
                if "permute" in info:
                    style_str = "draw=gray, dashed" if num_ctrl_system != 0 else "draw=none"
                    gate_prefix += r',style={' + style_str + r'}'
                gate_command = fr"\gate[{gate_prefix}]{{{gate_name}}}"
            
            self._code[min_idx].append(gate_command)
            for idx in range(min_idx + 1, max_idx + 1):
                self._code[idx].append(r"{}" if idx in system_idx else _empty_through)
    
    def _append_general_param_op(self, info: OperatorInfoType) -> None:
        r"""Append a general parameterized operator
        
        Note:
            Presentation rule is as follows:
            - If the style is 'detailed', the matrix is provided, 
            the number of acted parameters is greater than 3 and not controlled, show the matrix;
            - If the style is 'compact', just show the name;
            - Otherwise, show the name with parameters.
        """
        assert "param" in info, \
            "The parameterized operator must have a parameter"
        if "param_sharing" not in info:
            warnings.warn(
                f"info for param_sharing of operator {info['name']} is not provided, set to False", UserWarning)
            info["param_sharing"] = False
        
        if "tex" not in info:
            info["tex"] = rf"\operatorname{{{info['name']}}}"
        num_ctrl_system, control_label = info.get("num_ctrl_system", 0), info.get("label", None)
        
        if self.style == 'compact' or info['name'] == "universal":
            show_matrix = False
            show_param = False
        elif (self.style == 'detailed' and "matrix" in info and 
            info['param'].shape[-1] > 3 and num_ctrl_system == 0):
            show_matrix = True
            show_param = False
        else:
            show_matrix = False
            show_param = True
        
        for op_idx, system_idx in enumerate(info["system_idx"]):
            self._fill_empty(system_idx)
            min_idx, max_idx = self._add_ctrl(system_idx, num_ctrl_system, control_label)
            num_across_system = max_idx - min_idx + 1
            
            if show_matrix:
                gate_name = _format_data(info['matrix'][op_idx], self.decimals)
            elif show_param:
                param = info['param'][0, 0] if info['param_sharing'] else info['param'][op_idx, 0]
                param_str = _format_data(param, self.decimals)
                gate_name = info['tex'] + '(' + param_str + ')'
            else:
                gate_name = info['tex']
            
            gate_prefix = str(num_across_system)
            if show_param and self.style == 'detailed':
                param_str = _format_data((param / np.pi) % 2, self.decimals)
                param_str = ', '.join([rf"{num}\pi" for num in param_str.split(', ')]).replace(r'\ldots\pi', '\ldots')
                gate_prefix += r',label style={label={[gray]below:$\scriptscriptstyle ' + param_str + r' $}}'
            gate_command = r"\gate[" + gate_prefix + r']{' +  gate_name + r'}'
            
            self._code[min_idx].append(gate_command)
            for idx in range(min_idx + 1, max_idx + 1):
                self._code[idx].append(r"{}" if idx in system_idx else self.__empty_through)
    
    def _append_measure_op(self, info: OperatorInfoType) -> None:
        r"""Append a measurement operator
        """
        system_idx = info["system_idx"]
        self._fill_empty(system_idx)
        
        if (collapse_label := info.get("label", None)):
            for i, idx in enumerate(system_idx):
                self._code[idx].append(r"\measuretab{" + collapse_label[i] + r"}")
                self._code[idx].append(r"\wireoverride{n}")
        else:
            min_idx, max_idx = min(system_idx), max(system_idx)
            num_across_system = max_idx - min_idx + 1
            measure_command = r"\meter[" + str(num_across_system) + r"]{" + str(info.get('label', '')) + "}"
            self._code[min_idx].append(measure_command)
            for idx in range(min_idx, max_idx + 1):
                self._code[idx].append(r"{}" if idx in system_idx else self.__empty_through)
                
    def _append_locc_op(self, info: OperatorInfoType) -> None:
        r"""Append a local operation
        """
        system_idx, num_measure_system = info["system_idx"], info["num_ctrl_system"]
        measure_idx = system_idx[:num_measure_system]
        self._fill_empty(measure_idx)
        
        label = info["label"]
        if len(measure_idx) == 1:
            self._code[measure_idx[0]].append(r"\meter{}")
            self._code[measure_idx[0]].append(r"\push{" + label + r"} \wireoverride{c}")
        else:
            for i, idx in enumerate(measure_idx):
                self._code[idx].append(r"\meter{}")
                self._code[idx].append(r"\push{{" + label + r"}_" + str(i) + r"} \wireoverride{c}")
        self._fill_empty(system_idx, classical_idx=measure_idx)
        
        min_idx, max_idx = self._add_ctrl(system_idx, num_measure_system, ['1'] * num_measure_system, classical=True)
        for idx in measure_idx:
            self._code[idx][-1] += r"\wireoverride{c}"
        num_across_system = max_idx - min_idx + 1
        gate_name = r'{' + info["tex"] + r"}^{(" + label + r")}"
        gate_command = r"\gate[" + str(num_across_system) + r']{' + gate_name + r'}'
        self._code[min_idx].append(gate_command)
        for idx in range(min_idx + 1, max_idx + 1):
            self._code[idx].append(r"{}" if idx in system_idx else self.__empty_through)
                
    def append(self, info: OperatorInfoType) -> None:
        r"""Append an operator to the current circuit code
        
        Args:
            info: The information of the operator to append
        
        """
        if info['name'] == "measure":
            self._append_measure_op(info)
        elif info['name'] == "locc":
            self._append_locc_op(info)
        elif "param" in info:
            self._append_general_param_op(info)
        else:
            self._append_general_op(info)

    def draw_layer(self, list_info: List[OperatorInfoType], 
                   name: str, depth: int = 1) -> 'OperatorListDrawer':
        r"""Give the drawer for a layer, i.e., a list of operators
        
        Args:
            list_info: The list of operators to draw
            name: The name of the layer.
            depth: The depth of the layer. Defaults to 1.
        
        Note:
            if name is given, the layer name will be added to the code by
            - command '\gate' with layer name if the style is 'compact';
            - command '\gate' with matrix if the style is 'detailed' and matrix is given.
            - command '\gategroup' with name above if the style is 'standard';
        
        """
        assert depth > 0, \
                f"Layer must have non-zero depth: received {depth}"
        num_op_per_layer, num_remain_layer = divmod(len(list_info), depth)
        layer_style = self.style

        drawer = OperatorListDrawer(layer_style, self.decimals)

        if layer_style == 'compact':
            list_info, list_end_info = list_info[:num_op_per_layer], list_info[-num_remain_layer:]
        for info in list_info:
            drawer.append(info)

        min_idx, max_idx = min(drawer._code.keys()), max(drawer._code.keys())
        label_pos = 'above' if min_idx == 0 else 'below'

        drawer._code[min_idx][0] += (r'\gategroup[' + str(len(drawer._code)) + 
                                     r',steps=' + str(len(drawer._code[max_idx])) + 
                                     r',style={inner sep=4pt,dashed,label={' + label_pos + 
                                     r':{' + name + r'}}}]{}')
        drawer.fill_all()
        
        if layer_style == 'compact' and num_remain_layer > 0:
            for info in list_end_info:
                drawer.append(info)
            drawer.fill_all()
        return drawer



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The source file of the oracle class and the control oracle class.
"""

import math
from functools import partial
from typing import Callable, Dict, Iterable, List, Optional, Union

import matplotlib
import torch

from ...core import OperatorInfoType, State, get_device, get_dtype, utils
from ...core.intrinsic import _int_to_digit
from ...database.set import gell_mann
from .base import Gate, ParamGate
from .visual import _c_oracle_like_display, _oracle_like_display


class Oracle(Gate):
    r"""An oracle as a gate.

    Args:
        oracle: Unitary oracle to be implemented.
        system_idx: Indices of the systems on which the gates are applied.
        acted_system_dim: dimension of systems that this gate acts on. Can be a list of system dimensions 
            or an int representing the dimension of all systems. Defaults to be qubit case.
    """
    def __init__(
            self, oracle: torch.Tensor, system_idx: Union[List[int], int],
            acted_system_dim: Union[List[int], int] = 2, gate_info: Dict = None,
    ):
        default_gate_info = {
            "name": "oracle",
            "tex": r'\text{O}',
            "api": "oracle",
            'plot_width': 0.6,
        }
        default_gate_info.update(gate_info or {})
        super().__init__(oracle, system_idx, acted_system_dim, gate_info=default_gate_info)

    @property
    def info(self) -> OperatorInfoType:
        r"""Information of this gate.
        """
        oracle = self.matrix
        info = super().info
        info.update({
            'matrix': oracle,
        })
        return info
    
    def display_in_circuit(self, ax: matplotlib.axes.Axes, x: float,) -> float:
        return _oracle_like_display(self, ax, x)


class ControlOracle(Gate):
    r"""A controlled oracle as a gate.
    
    Args:
        oracle: Unitary oracle to be implemented.
        system_idx: Indices of the systems on which the gates are applied. The first element in the list is the control system, 
            represented by a int or a list of ints, and the remaining elements represent the oracle system.
        index: control index that activates the matrix.
        acted_system_dim: dimension of systems that this gate acts on. Can be a list of system dimensions 
            or an int representing the dimension of all systems. Defaults to be qubit case.
        gate_info: information of this gate that will be placed into the gate history or plotted by a Circuit. 
            Defaults to ``None``.
    
    """
    def __init__(
        self, oracle: torch.Tensor, system_idx: List[Union[List[int], int]], index: int, 
        acted_system_dim: Union[List[int], int] = 2, gate_info: Optional[Dict] = None
    ) -> None:
        ctrl_system_idx = [system_idx[0]] if isinstance(system_idx[0], int) else system_idx[0]
        
        num_ctrl_system = len(ctrl_system_idx)
        ctrl_system_dim = [acted_system_dim] * num_ctrl_system if isinstance(acted_system_dim, int) else acted_system_dim[:num_ctrl_system]
        default_gate_info = {
            "name": "ctrl-oracle",
            "tex": r'\text{O}',
            "api": "control_oracle",
            "num_ctrl_system": len(ctrl_system_idx),
            "label": _int_to_digit(index, ctrl_system_dim).zfill(num_ctrl_system),
            'plot_width': 0.6,
        }
        default_gate_info.update(gate_info or {})
        super().__init__(None, ctrl_system_idx + system_idx[1:], acted_system_dim, gate_info=default_gate_info)
        
        self.__system_idx = [ctrl_system_idx] + system_idx[1:]
        self.__apply_matrix = oracle.to(dtype=self.dtype, device=self.device)
        self.__index = index
        
    @property
    def matrix(self) -> torch.Tensor:
        ctrl_dim = math.prod(self.system_dim[:len(self.__system_idx[0])])
        index = self.__index
        
        proj = torch.zeros([ctrl_dim, ctrl_dim])
        proj[index, index] = 1
        
        matrix = self.__apply_matrix
        _eye = torch.eye(matrix.shape[-1]).expand_as(matrix)
        return utils.linalg._kron(proj, matrix) + utils.linalg._kron(torch.eye(ctrl_dim) - proj, _eye)
    
    def forward(self, state: State) -> State:
        state = state.clone()
        state._evolve_ctrl(self.__apply_matrix, self.__index, self.__system_idx)
        return state
    
    def display_in_circuit(self, ax: matplotlib.axes.Axes, x: float,) -> float:
        return _c_oracle_like_display(self, ax, x)


class ParamOracle(ParamGate):
    r"""An parameterized oracle as a gate

    Args:
        generator: function that generates the oracle.
        system_idx: indices of the system on which this gate acts on.
        param: input parameters of quantum parameterized gates. Defaults to ``None`` i.e. randomized.
        num_acted_param: the number of parameters required for a single operation.
        acted_system_dim: dimension of systems that this gate acts on. Can be a list of system dimensions 
            or an int representing the dimension of all systems. Defaults to be qubit case.
        gate_info: information of this gate that will be placed into the gate history or plotted by a Circuit. 
            Defaults to ``None``.
        support_batch: whether the generator supports batched input. Defaults to ``False``.

    """
    def __init__(
        self, generator: Callable[[torch.Tensor], torch.Tensor], system_idx: Union[List[int], int],
        param: Union[torch.Tensor, float, List[float]] = None, num_acted_param: int = 1,
        acted_system_dim: Union[List[int], int] = 2, gate_info: Dict = None, support_batch: bool = False,
    ):
        default_gate_info = {
            "name": "ctrl-param-oracle",
            "tex": r'P',
            "param_sharing": False,
            "api": "param_oracle",
            "kwargs": {"generator": generator},
            'plot_width': 0.6,
        }
        default_gate_info.update(gate_info or {})
        super().__init__(generator, param, num_acted_param, False, system_idx, acted_system_dim, True, default_gate_info, support_batch)


class ControlParamOracle(ParamGate):
    r"""An parameterized oracle as a gate

    Args:
        generator: function that generates the oracle.
        system_idx: Indices of the systems on which the gates are applied. The first element in the list is the control system, 
            represented by a int or a list of ints, and the remaining elements represent the oracle system.
        index: control index that activates the matrix.
        param: input parameters of quantum parameterized gates. Defaults to ``None`` i.e. randomized.
        num_acted_param: the number of parameters required for a single operation.
        acted_system_dim: dimension of systems that this gate acts on. Can be a list of system dimensions 
            or an int representing the dimension of all systems. Defaults to be qubit case.
        gate_info: information of this gate that will be placed into the gate history or plotted by a Circuit. 
            Defaults to ``None``.
        support_batch: whether the generator supports batched input. Defaults to ``False``.

    """
    def __init__(
        self, generator: Callable[[torch.Tensor], torch.Tensor], system_idx: Union[List[int], int], index: int, 
        param: Union[torch.Tensor, float, List[float]] = None, num_acted_param: int = 1,
        acted_system_dim: Union[List[int], int] = 2, gate_info: Dict = None, support_batch: bool = False,
    ):
        ctrl_system_idx = [system_idx[0]] if isinstance(system_idx[0], int) else system_idx[0]
        
        num_ctrl_system = len(ctrl_system_idx)
        ctrl_system_dim = [acted_system_dim] * num_ctrl_system if isinstance(acted_system_dim, int) else acted_system_dim[:num_ctrl_system]
        default_gate_info = {
            "name": "ctrl-param-oracle",
            "tex": r'P',
            "param_sharing": False,
            "api": "param_oracle",
            "num_ctrl_system": len(ctrl_system_idx),
            "label": _int_to_digit(index, ctrl_system_dim).zfill(num_ctrl_system),
            'plot_width': 0.6,
        }
        default_gate_info.update(gate_info or {})
        super().__init__(generator, param, num_acted_param, False, ctrl_system_idx + system_idx[1:], 
                         acted_system_dim, True, default_gate_info, support_batch)
        
        self.__system_idx = [ctrl_system_idx] + system_idx[1:]
        self.__index = index
        
    @property
    def matrix(self) -> torch.Tensor:
        ctrl_dim = math.prod(self.system_dim[:len(self.__system_idx[0])])
        index = self.__index
        
        proj = torch.zeros([ctrl_dim, ctrl_dim])
        proj[index, index] = 1
        
        matrix = super().matrix
        _eye = torch.eye(matrix.shape[-1]).expand_as(matrix)
        return utils.linalg._kron(proj, matrix) + utils.linalg._kron(torch.eye(ctrl_dim) - proj, _eye)
    
    def forward(self, state: State) -> State:
        state = state.clone()
        state._evolve_ctrl(super().matrix, self.__index, self.__system_idx)
        return state


def _universal_matrix(param: torch.Tensor, bases: torch.Tensor) -> torch.Tensor:
    r"""Generate a universal matrix with the given parameters and bases.
    """
    h = torch.sum(torch.mul(param.view([-1, 1, 1]), bases), dim=-3)
    return torch.matrix_exp(1j * h)

class UniversalQudits(ParamGate):
    r"""A collection of universal qudit gates. One of such a gate requires :math:`d^2 - 1` parameters.

    Args:
        system_idx: Indices of the qubits on which the gates are applied. Defaults to the first qubit.
        param: Parameters of the gates. Defaults to ``None``.
        param_sharing: Whether gates in the same layer share a parameter. Defaults to ``False``.
        acted_system_dim: dimension of systems that this gate acts on. Can be a list of system dimensions 
            or an int representing the dimension of all systems. Defaults to be qubit case.
    """
    def __init__(
        self, system_idx: Optional[Union[Iterable[int], str]], acted_system_dim: Iterable[int],
        param: Optional[Union[torch.Tensor, float]] = None,
        param_sharing: Optional[bool] = False,

    ):
        assert not isinstance(acted_system_dim, int), \
            f"system dimensions for UniversalQudits cannot be a integer: received {acted_system_dim}"

        dim = math.prod(acted_system_dim)
        generator = gell_mann(dim).to(get_device(), dtype=get_dtype())
        matrix_func = partial(utils.matrix._param_generator, generator=generator)

        gate_info = {
            "name": "universal",
            "tex": r'\operatorname{UNI}_{' + str(dim) + r'}',
            "api": "universal_qudits",
            "param_sharing": param_sharing,
            'plot_width': 0.8,
        }
        super().__init__(
            matrix_func, param, dim ** 2 - 1, param_sharing, system_idx, acted_system_dim,
            check_legality=False, gate_info=gate_info)

    def display_in_circuit(self, ax: matplotlib.axes.Axes, x: float, ) -> float:
        return _oracle_like_display(self, ax, x)


class Permutation(Gate):
    r"""permutation matrix as a gate.

    Args:
        perm: A list representing the permutation of subsystems.
            For example, [1, 0, 2] swaps the first two subsystems.
        system_idx: Indices of the systems on which the gates are applied.
        acted_system_dim: dimension of systems that this gate acts on. Can be a list of system dimensions 
            or an int representing the dimension of all systems. Defaults to be qubit case.
    
    """
    def __init__(
            self, perm: List[int], system_idx: List[Union[List[int], int]],
            acted_system_dim: Union[List[int], int] = 2
        ):
        gate_info = {
            "name": "permute",
            "api": "permute",
            "permute": perm,
            'plot_width': 0.2,
        }
        
        self.perm = perm
        super().__init__(None, system_idx, acted_system_dim, check_legality=False, gate_info=gate_info)
    
    @property
    def matrix(self) -> torch.Tensor:
        return utils.matrix._permutation(self.perm, self.system_dim)
    
    def forward(self, state: State) -> State:
        perm, system_idx = self.perm, self.system_idx[0]
        key_map = {system_idx[i]: system_idx[perm[i]] for i in range(len(system_idx))}
        target_seq = [(key_map[idx] if idx in system_idx else idx) for idx in range(state.num_systems)]
        return state.permute(target_seq)



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The source file of the classes for common quantum channels.
"""

from typing import Iterable, Union

import numpy as np
import torch

from ...core import State
from ...database.representation import (amplitude_damping_kraus,
                                        bit_flip_kraus, bit_phase_flip_kraus,
                                        depolarizing_kraus,
                                        generalized_amplitude_damping_kraus,
                                        generalized_depolarizing_kraus,
                                        pauli_kraus, phase_damping_kraus,
                                        phase_flip_kraus, replacement_choi,
                                        reset_kraus, thermal_relaxation_kraus)
from .base import Channel


class BitFlip(Channel):
    r"""A collection of bit flip channels.

    Such a channel's Kraus operators are

    .. math::

        E_0 = \sqrt{1-p} I,
        E_1 = \sqrt{p} X.

    Args:
        prob: Probability of a bit flip. Its value should be in the range :math:`[0, 1]`.
        qubits_idx: Indices of the qubits on which the channels act. Defaults to list(range(# of acted qubits))
    """
    def __init__(
            self, prob: Union[torch.Tensor, float],
            qubits_idx: Union[Iterable[int], int, str] = None
    ):
        p = np.round(prob.item() if isinstance(prob, torch.Tensor) else prob, decimals=2)
        channel_info = {
            'name': 'bit_flip',
            'tex': r'\mathcal{E}_{p = ' + str(p) + r'}^{\textrm{\tiny{(BF)}}}',
            'api': 'bit_flip',
            'kwargs': {'prob': prob}
        }
        super().__init__('kraus', bit_flip_kraus(prob), qubits_idx, 
                         acted_system_dim=2, check_legality=False, channel_info=channel_info)


class PhaseFlip(Channel):
    r"""A collection of phase flip channels.

    Such a channel's Kraus operators are

    .. math::

        E_0 = \sqrt{1 - p} I,
        E_1 = \sqrt{p} Z.

    Args:
        prob: Probability of a phase flip. Its value should be in the range :math:`[0, 1]`.
        qubits_idx: Indices of the qubits on which the channels act. Defaults to list(range(# of acted qubits))
    """
    def __init__(
            self, prob: Union[torch.Tensor, float],
            qubits_idx: Union[Iterable[int], int, str] = None
    ):
        p = np.round(prob.item() if isinstance(prob, torch.Tensor) else prob, decimals=2)
        channel_info = {
            'name': 'phase_flip',
            'tex': r'\mathcal{E}_{p = ' + str(p) + r'}^{\textrm{\tiny{(PF)}}}',
            'api': 'phase_flip',
            'kwargs': {'prob': prob}
        }
        super().__init__('kraus', phase_flip_kraus(prob), qubits_idx, 
                         acted_system_dim=2, check_legality=False, channel_info=channel_info)


class BitPhaseFlip(Channel):
    r"""A collection of bit phase flip channels.

    Such a channel's Kraus operators are

    .. math::

        E_0 = \sqrt{1 - p} I,
        E_1 = \sqrt{p} Y.

    Args:
        prob: Probability of a bit phase flip. Its value should be in the range :math:`[0, 1]`.
        qubits_idx: Indices of the qubits on which the channels act. Defaults to list(range(# of acted qubits))
    """
    def __init__(
            self, prob: Union[torch.Tensor, float],
            qubits_idx: Union[Iterable[int], int, str] = None
    ):
        p = np.round(prob.item() if isinstance(prob, torch.Tensor) else prob, decimals=2)
        channel_info = {
            'name': 'bit_phase_flip',
            'tex': r'\mathcal{E}_{p = ' + str(p) + r'}^{\textrm{\tiny{(BPF)}}}',
            'api': 'bit_phase_flip',
            'kwargs': {'prob': prob}
        }
        super().__init__('kraus', bit_phase_flip_kraus(prob), qubits_idx, 
                         acted_system_dim=2, check_legality=False, channel_info=channel_info)


class AmplitudeDamping(Channel):
    r"""A collection of amplitude damping channels.

    Such a channel's Kraus operators are

    .. math::

        E_0 =
        \begin{bmatrix}
            1 & 0 \\
            0 & \sqrt{1-\gamma}
        \end{bmatrix},
        E_1 =
        \begin{bmatrix}
            0 & \sqrt{\gamma} \\
            0 & 0
        \end{bmatrix}.

    Args:
        gamma: Damping probability. Its value should be in the range :math:`[0, 1]`.
        qubits_idx: Indices of the qubits on which the channels act. Defaults to list(range(# of acted qubits))
    """
    def __init__(
            self, gamma: Union[torch.Tensor, float],
            qubits_idx: Union[Iterable[int], int, str] = None
    ):
        g = np.round(gamma.item() if isinstance(gamma, torch.Tensor) else gamma, decimals=2)
        channel_info = {
            'name': 'amplitude_damping',
            'tex': r'\mathcal{E}_{\gamma = ' + str(g) + r'}^{\textrm{\tiny{(AD)}}}',
            'api': 'amplitude_damping',
            'kwargs': {'gamma': gamma}
        }
        super().__init__('kraus', amplitude_damping_kraus(gamma), qubits_idx, 
                         acted_system_dim=2, check_legality=False, channel_info=channel_info)


class GeneralizedAmplitudeDamping(Channel):
    r"""A collection of generalized amplitude damping channels.

    Such a channel's Kraus operators are

    .. math::

        E_0 = \sqrt{p}
        \begin{bmatrix}
            1 & 0 \\
            0 & \sqrt{1-\gamma}
        \end{bmatrix},
        E_1 = \sqrt{p} \begin{bmatrix} 0 & \sqrt{\gamma} \\ 0 & 0 \end{bmatrix},\\
        E_2 = \sqrt{1-p} \begin{bmatrix} \sqrt{1-\gamma} & 0 \\ 0 & 1 \end{bmatrix},
        E_3 = \sqrt{1-p} \begin{bmatrix} 0 & 0 \\ \sqrt{\gamma} & 0 \end{bmatrix}.

    Args:
        gamma: Damping probability. Its value should be in the range :math:`[0, 1]`.
        prob: Excitation probability. Its value should be in the range :math:`[0, 1]`.
        qubits_idx: Indices of the qubits on which the channels act. Defaults to list(range(# of acted qubits))
    """
    def __init__(
            self, gamma: Union[torch.Tensor, float], prob: Union[torch.Tensor, float],
            qubits_idx: Union[Iterable[int], int, str] = None
    ):
        g = np.round(gamma.item() if isinstance(gamma, torch.Tensor) else gamma, decimals=2)
        p = np.round(prob.item() if isinstance(prob, torch.Tensor) else prob, decimals=2)
        channel_info = {
            'name': 'generalized_amplitude_damping',
            'tex': r'\mathcal{E}_{\gamma = ' + str(g) + r', p = ' + str(p) + r'}^{\textrm{\tiny{(GAD)}}}',
            'api': 'generalized_amplitude_damping',
            'kwargs': {'gamma': gamma, 'prob': prob}
        }
        super().__init__(
            'kraus', generalized_amplitude_damping_kraus(gamma, prob), qubits_idx, 
            acted_system_dim=2, check_legality=False, channel_info=channel_info)


class PhaseDamping(Channel):
    r"""A collection of phase damping channels.

    Such a channel's Kraus operators are

    .. math::

        E_0 =
        \begin{bmatrix}
            1 & 0 \\
            0 & \sqrt{1-\gamma}
        \end{bmatrix},
        E_1 =
        \begin{bmatrix}
            0 & 0 \\
            0 & \sqrt{\gamma}
        \end{bmatrix}.

    Args:
        gamma: Parameter of the phase damping channels. Its value should be in the range :math:`[0, 1]`.
        qubits_idx: Indices of the qubits on which the channels act. Defaults to list(range(# of acted qubits))
    """
    def __init__(
            self, gamma: Union[torch.Tensor, float],
            qubits_idx: Union[Iterable[int], int, str] = None
    ):
        g = np.round(gamma.item() if isinstance(gamma, torch.Tensor) else gamma, decimals=2)
        channel_info = {
            'name': 'phase_damping',
            'tex': r'\mathcal{E}_{\gamma = ' + str(g) + r'}^{\textrm{\tiny{(PD)}}}',
            'api': 'phase_damping',
            'kwargs': {'gamma': gamma}
        }
        super().__init__('kraus', phase_damping_kraus(gamma), qubits_idx, 
                         acted_system_dim=2, check_legality=False, channel_info=channel_info)


class Depolarizing(Channel):
    r"""A collection of depolarizing channels.

    Such a channel's Kraus operators are

    .. math::

        E_0 = \sqrt{1-3p/4} I,
        E_1 = \sqrt{p/4} X,
        E_2 = \sqrt{p/4} Y,
        E_3 = \sqrt{p/4} Z.

    Args:
        prob: Parameter of the depolarizing channels. Its value should be in the range :math:`[0, 1]`.
        qubits_idx: Indices of the qubits on which the channels act. Defaults to list(range(# of acted qubits))

    Note:
        The implementation logic for this feature has been updated.
        The current version refers to formula (8.102) in Quantum Computation and Quantum Information 10th
        edition by M.A.Nielsen and I.L.Chuang.
        Reference: Nielsen, M., & Chuang, I. (2010). Quantum Computation and Quantum Information: 10th
        Anniversary Edition. Cambridge: Cambridge University Press. doi:10.1017/CBO9780511976667
    """
    def __init__(
            self, prob: Union[torch.Tensor, float],
            qubits_idx: Union[Iterable[int], int, str] = None
    ):
        p = np.round(prob.item() if isinstance(prob, torch.Tensor) else prob, decimals=2)
        channel_info = {
            'name': 'depolarizing',
            'tex': r'\mathcal{D}_{p = ' + str(p) + r'}',
            'api': 'depolarizing',
            'kwargs': {'prob': prob}
        }
        super().__init__('kraus', depolarizing_kraus(prob), qubits_idx, 
                         acted_system_dim=2, check_legality=False, channel_info=channel_info)


class GeneralizedDepolarizing(Channel):
    r"""A generalized depolarizing channel.

    Such a channel's Kraus operators are

    .. math::

        E_0 = \sqrt{1-(D - 1)p/D} I, \text{ where } D = 4^n, \\
        E_k = \sqrt{p/D} \sigma_k, \text{ for } 0 < k < D.

    Args:
        prob: probability :math:`p`. Its value should be in the range :math:`[0, 1]`.
        qubits_idx: Indices of the qubits on which the channels act, the length of which is :math:`n`.
            Defaults to be ``None``.
        num_qubits: Total number of qubits. Defaults to ``None``.
    """
    def __init__(
            self, prob: Union[torch.Tensor, float],
            qubits_idx: Union[Iterable[int], int, str], num_qubits: int = None
    ):
        p = np.round(prob.item() if isinstance(prob, torch.Tensor) else prob, decimals=2)
        channel_info = {
            'name': 'generalized_depolarizing',
            'tex': r'\mathcal{D}_{p = ' + str(p) + r'}',
            'api': 'generalized_depolarizing',
            'kwargs': {'prob': prob}
        }
        if qubits_idx in ['full', 'cycle']:
            raise NotImplementedError(
                'The generalized depolarizing channel should act on all qubits of the system')
        num_acted_qubits = np.size(np.array(qubits_idx))
        super().__init__(
            'kraus', generalized_depolarizing_kraus(prob, num_acted_qubits), qubits_idx, 
            acted_system_dim=2, check_legality=False, channel_info=channel_info)


class PauliChannel(Channel):
    r"""A collection of Pauli channels.

    Args:
        prob: Probabilities corresponding to the Pauli X, Y, and Z operators. Each value should be in the
            range :math:`[0, 1]`.
        qubits_idx: Indices of the qubits on which the channels act. Defaults to list(range(# of acted qubits))

    Note:
        The sum of three input probabilities should be less than or equal to 1.
    """
    def __init__(
            self, prob: Union[torch.Tensor, Iterable[float]],
            qubits_idx: Union[Iterable[int], int, str] = None
    ):
        channel_info = {
            'name': 'pauli_channel',
            'tex': r'\mathcal{N}',
            'api': 'pauli_channel',
            'kwargs': {'prob': prob}
        }
        super().__init__('kraus', pauli_kraus(prob), qubits_idx, 
                         acted_system_dim=2, check_legality=False, channel_info=channel_info)


class ResetChannel(Channel):
    r"""A collection of reset channels.

    Such a channel reset the state to :math:`|0\rangle` with a probability of p and to :math:`|1\rangle` with
    a probability of q. Its Kraus operators are

    .. math::

        E_0 =
        \begin{bmatrix}
            \sqrt{p} & 0 \\
            0 & 0
        \end{bmatrix},
        E_1 =
        \begin{bmatrix}
            0 & \sqrt{p} \\
            0 & 0
        \end{bmatrix},\\
        E_2 =
        \begin{bmatrix}
            0 & 0 \\
            \sqrt{q} & 0
        \end{bmatrix},
        E_3 =
        \begin{bmatrix}
            0 & 0 \\
            0 & \sqrt{q}
        \end{bmatrix},\\
        E_4 = \sqrt{1-p-q} I.

    Args:
        prob: Probabilities of resetting to :math:`|0\rangle` and to :math:`|1\rangle`. Each value should be
            in the range :math:`[0, 1]`.
        qubits_idx: Indices of the qubits on which the channels act. Defaults to list(range(# of acted qubits))

    Note:
        The sum of two input probabilities should be less than or equal to 1.
    """
    def __init__(
            self, prob: Union[torch.Tensor, Iterable[float]],
            qubits_idx: Union[Iterable[int], int, str] = None
    ):
        channel_info = {
            'name': 'reset_channel',
            'tex': r'\mathcal{N}',
            'api': 'reset_channel',
            'kwargs': {'prob': prob}
        }
        super().__init__('kraus', reset_kraus(prob), qubits_idx, 
                         acted_system_dim=2, check_legality=False, channel_info=channel_info)


class ThermalRelaxation(Channel):
    r"""A collection of thermal relaxation channels.

    Such a channel simulates the mixture of the :math:`T_1` and the :math:`T_2` processes on superconducting devices.

    Args:
        const_t: :math:`T_1` and :math:`T_2` relaxation time in microseconds.
        exec_time: Quantum gate execution time in the process of relaxation in nanoseconds.
        qubits_idx: Indices of the qubits on which the channels act. Defaults to list(range(# of acted qubits))

    Note:
        Relaxation time must satisfy :math:`T_2 \le T_1`. For reference please see https://arxiv.org/abs/2101.02109.
    """
    def __init__(
            self, const_t: Union[torch.Tensor, Iterable[float]], exec_time: Union[torch.Tensor, float],
            qubits_idx: Union[Iterable[int], int, str] = None
    ):
        channel_info = {
            'name': 'thermal_relaxation',
            'tex': r'\mathcal{N}',
            'api': 'thermal_relaxation',
            'kwargs': {'const_t': const_t, 'exec_time': exec_time}
        }
        super().__init__(
            'kraus', thermal_relaxation_kraus(const_t, exec_time),
            qubits_idx, acted_system_dim=2, check_legality=False, channel_info=channel_info)


class ReplacementChannel(Channel):
    r"""A collection of quantum replacement channels.

    For a quantum state :math:`\sigma`, the corresponding replacement channel :math:`R` is defined as

    .. math::

        R(\rho) = \text{tr}(\rho)\sigma

    Args:
        sigma: The state to be replaced.
        system_idx: Indices of the qubits on which the channels act, the length of which is :math:`n`.
            Defaults to list(range(# of acted qubits)).
    """
    def __init__(
            self, sigma: State,
            system_idx: Union[Iterable[int], int, str] = None
    ):
        channel_info = {
            'name': 'replace_channel',
            'tex': r'\mathcal{N}',
        }
        super().__init__('choi', replacement_choi(sigma.density_matrix), system_idx, 
                         acted_system_dim=sigma.dim, channel_info=channel_info)
